# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
"""
EdgeMiner
=========

A module for detecting linked edges.

The complementary module ezdxf.edgesmith can create entities from the output of this 
module.

Terminology
-----------

I try to use the terminology of `Graph Theory`_ but there are differences where I think
a different term is better suited for this module like loop for cycle.

Edge (in this module)
    Edge is an immutable class:
        - unique id
        - 3D start point (vertex)
        - 3D end point (vertex)
        - optional length
        - optional payload (arbitrary data)

    The geometry of an edge is not known.
    Intersection points of edges are not known and cannot be calculated.

Vertex
    A connection point of two or more edges. 
    The degree of a vertex is the number of connected edges.

Leaf
    A leaf is a vertex of degree 1.
    A leaf is a loose end of an edge, which is not connected to other edges.

Junction
    A junction is a vertex of degree greater 2.
    A junction has more than two adjacent edges.
    A junction is an ambiguity when searching for open chains or closed loops.
    Graph Theory: multiple adjacency

Chain
    A chain has sequential connected edges. 
    The end point of an edge is connected to the start point of the following edge. 
    A chain has unique edges, each edge appears only once in the chain.
    A chain can contain vertices of degree greater 2.
    A solitary edge is also a chain.
    Chains are represented as Sequence[Edge].
    Graph Theory: Trail - no edge is repeated, vertex is repeated

Simple Chain (special to this module)
    A simple chain contains only vertices of degree 2, except the start- and end vertex.
    The start- and end vertices are leafs (degree of 1) or junctions (degree greater 2).
    
Open Chain
    An open chain is a chain which starts and ends with at leaf.
    A solitary edge is also an open chain.
    Graph Theory: Path - no edge is repeated, no vertex is repeated, endings not connected

Loop
    A loop is a simple chain with connected start- and end vertices.
    A loop has two or more edges.
    A loop contains only vertices of degree 2.
    Graph Theory: Cycle - no edge is repeated, no vertex is repeated, endings connected; 
    a loop in Graph Theory is something different

Network
    A network has two or more edges that are directly and indirectly connected. 
    The edges in a network have no order.
    A network can contain vertices of degree greater 2 (junctions).
    A solitary edge is not a network. 
    A chain with two or more edges is a network. 
    Networks are represented as Sequence[Edge].
    Graph Theory: multigraph; a network in Graph Theory is something different

Gap Tolerance
    Maximum vertex distance to consider two edges as connected

Forward Connection
    An edge is forward connected when the end point of the edge is connected to the 
    start point of the following edge.

.. important::
    
    THIS MODULE IS WORK IN PROGRESS (ALPHA VERSION), EVERYTHING CAN CHANGE UNTIL 
    THE RELEASE IN EZDXF V1.4.

.. _Graph Theory: https://en.wikipedia.org/wiki/Glossary_of_graph_theory
.. _GeeksForGeeks: https://www.geeksforgeeks.org/graph-data-structure-and-algorithms/?ref=shm

"""
from __future__ import annotations
from typing import Any, Sequence, Iterator, Iterable, Dict, Tuple, NamedTuple, Callable
from typing_extensions import Self, TypeAlias
from collections import Counter
import functools
import time
import math

from ezdxf.math import UVec, Vec2, Vec3, distance_point_line_3d
from ezdxf.math import rtree


__all__ = [
    "Deposit",
    "Edge",
    "TimeoutError",
    "find_all_loops",
    "find_all_open_chains",
    "find_all_sequential_chains",
    "find_all_simple_chains",
    "find_loop",
    "find_loop_by_edge",
    "find_sequential_chain",
    "flatten",
    "is_chain",
    "is_loop",
    "length",
    "longest_chain",
    "reverse_chain",
    "shortest_chain",
    "subtract_edges",
    "unique_chains",
]
GAP_TOL = 1e-9
ABS_TOL = 1e-9
TIMEOUT = 60.0  # in seconds


class TimeoutError(Exception):  # noqa
    """
    Attributes:
        solutions: solutions found until time out occur

    """

    def __init__(self, msg: str, solutions: Sequence[Sequence[Edge]] = tuple()) -> None:
        super().__init__(msg)
        self.solutions = solutions


class Watchdog:
    def __init__(self, timeout=TIMEOUT) -> None:
        self.timeout: float = timeout
        self.start_time: float = time.perf_counter()

    def start(self, timeout: float):
        self.timeout = timeout
        self.start_time = time.perf_counter()

    @property
    def has_timed_out(self) -> bool:
        return time.perf_counter() - self.start_time > self.timeout


class Edge(NamedTuple):
    """Represents an immutable edge.

    An edge can represent any linear curve (line, elliptical arc, spline, ...), but it
    does not know this shape; only the start and end vertices are stored, not the shape
    itself.  Therefore, the length of the edge must be specified if the length
    calculation for a sequence of edges should be possible.  Intersection points between
    edges are not known and cannot be calculated.

    .. Important::

        Use only the :func:`make_edge` function to create new edges to get unique ids!

    Attributes:
        id: unique id
        start (Vec3): start vertex
        end (Vec3): end vertex
        is_reverse: flag to indicate that the edge is reversed compared to its initial state
        length: length of the edge, default is 1.0
        payload: arbitrary data attached to the edge, default is ``None``

    """

    id: int
    start: Vec3
    end: Vec3
    is_reverse: bool = False
    length: float = 1.0
    payload: Any = None

    def __eq__(self, other) -> bool:
        """Return ``True`` if the ids of the edges are equal.

        .. important::

            An edge is equal to its reversed copy!

        """
        if isinstance(other, Edge):
            return self.id == other.id
        return False

    def __repr__(self) -> str:
        payload = self.payload
        if payload is None:
            content = str(self.id)
        elif isinstance(payload, EdgeWrapper):
            content = "[" + (",".join(repr(e) for e in payload.edges)) + "]"
        else:
            content = str(payload)
        return f"Edge({content})"

    def __hash__(self) -> int:
        """The edge :attr:`id` is used as hash value.

        An Edge and its reversed edge have the same hash value and cannot both
        exist in the same :class:`set`.
        """
        return self.id

    def reversed(self) -> Self:
        """Returns a reversed copy.

        The reversed edge has the same :attr:`id` as the source edge, because they
        represent the same edge.
        """
        return self.__class__(  # noqa
            self.id,  # edge and its reversed edge must have the same id!
            self.end,
            self.start,
            not self.is_reverse,
            self.length,
            self.payload,
        )


def make_id_generator(start: int = 0) -> Callable[[], int]:
    next_edge_id = start

    def next_id() -> int:
        nonlocal next_edge_id
        next_edge_id += 1
        return next_edge_id

    return next_id


id_generator = make_id_generator()


def make_edge(
    start: UVec, end: UVec, length: float = -1.0, *, payload: Any = None
) -> Edge:
    """Creates a new :class:`Edge` with a unique id.

    Args:
        start: start point
        end: end point
        length: default is the distance between start and end
        payload: arbitrary data attached to the edge

    """
    start = Vec3(start)
    end = Vec3(end)
    if length < 0.0:
        length = start.distance(end)
    return Edge(id_generator(), start, end, False, length, payload)


def isclose(a: Vec3, b: Vec3, *, gap_tol=GAP_TOL) -> bool:
    """This function should be used to test whether two vertices are close to each other
    to get consistent results.
    """
    return a.distance(b) <= gap_tol


class Deposit:
    """The edge deposit stores all available edges for further searches.

    The edges and the search index are immutable after instantiation.
    The :attr:`gap_tol` attribute is mutable.

    Args:
        edges: sequence of :class:`Edge`
        gap_tol: maximum vertex distance to consider two edges as connected

    Attributes:
        gap_tol: maximum vertex distance to consider two edges as connected (mutable)

    """

    def __init__(self, edges: Sequence[Edge], *, gap_tol=GAP_TOL) -> None:
        self.gap_tol: float = gap_tol
        self._edges: Sequence[Edge] = type_check(edges)
        self._search_index = _SpatialSearchIndex(self._edges)

    @property
    def edges(self) -> Sequence[Edge]:
        """Sequence of edges stored in this deposit."""
        return self._edges

    def degree_counter(self) -> Counter[int]:
        """Returns a :class:`Counter` for the degree of all vertices.

        - :code:`Counter[degree]` returns the count of vertices of this degree.
        - :code:`Counter.keys()` returns all existing degrees in this deposit

        A new counter will be created for every method call! The :attr:`gap_tol`
        attribute is mutable and different gap tolerances may yield different results.

        """
        # no caching: result depends on gap_tol, which is mutable
        counter: Counter[int] = Counter()
        search = functools.partial(
            self._search_index.vertices_in_sphere, radius=self.gap_tol
        )
        for edge in self.edges:
            counter[len(search(edge.start))] += 1
            counter[len(search(edge.end))] += 1
        # remove duplicate counts:
        return Counter({k: v // k for k, v in counter.items()})

    @property
    def max_degree(self) -> int:
        """Returns the maximum degree of all vertices in this deposit."""
        return max(self.degree_counter().keys())

    def degree(self, vertex: UVec) -> int:
        """Returns the degree of the given vertex.

        - degree of 0: not in this deposit
        - degree of 1: one edge is connected to this vertex
        - degree of 2: two edges are connected to this vertex
        - degree of 3: three edges ... and so on


        Check if a vertex exist in a deposit::

            if deposit.degree(vertex): ...
        """
        return len(self._search_index.vertices_in_sphere(Vec3(vertex), self.gap_tol))

    def degrees(self, vertices: Iterable[UVec]) -> Sequence[int]:
        """Returns the degree of the given vertices."""
        search = functools.partial(
            self._search_index.vertices_in_sphere, radius=self.gap_tol
        )
        return tuple(len(search(vertex)) for vertex in Vec3.generate(vertices))

    def unique_vertices(self) -> set[Vec3]:
        """Returns all unique vertices from this deposit.

        Ignores vertices that are close to another vertex (within the range of gap_tol).
        It is not determined which of the close vertices is returned.

        e.g. if the vertices a, b are close together, you don't know if you get a or b,
        but it's guaranteed that you only get one of them
        """
        return filter_close_vertices(self._search_index.rtree, gap_tol=self.gap_tol)

    def edges_linked_to(self, vertex: UVec, radius: float = -1) -> Sequence[Edge]:
        """Returns all edges linked to `vertex` in range of `radius`.

        Args:
            vertex: 3D search location
            radius: search range, default radius is :attr:`Deposit.gap_tol`

        """
        if radius < 0:
            radius = self.gap_tol
        vertices = self._search_index.vertices_in_sphere(Vec3(vertex), radius)
        return tuple(v.edge for v in vertices)

    def find_nearest_edge(self, vertex: UVec) -> Edge | None:
        """Return the nearest edge to the given vertex.

        The distance is measured to the connection line from start to end of the edge.
        This is not correct for edges that represent arcs or splines.
        """

        def distance(edge: Edge) -> float:
            try:
                return distance_point_line_3d(vertex, edge.start, edge.end)
            except ZeroDivisionError:
                return edge.start.distance(vertex)

        vertex = Vec3(vertex)
        si = self._search_index
        nearest_vertex = si.nearest_vertex(vertex)
        edges = self.edges_linked_to(nearest_vertex)
        if edges:
            return min(edges, key=distance)
        return None

    def find_network(self, edge: Edge) -> set[Edge]:
        """Returns the network of all edges that are directly and indirectly linked to
        `edge`.  A network has two or more edges, a solitary edge is not a network.

        """

        def process(vertex: Vec3) -> None:
            linked_edges = set(self.edges_linked_to(vertex)) - network
            if linked_edges:
                network.update(linked_edges)
                todo.extend(linked_edges)

        todo: list[Edge] = [edge]
        network: set[Edge] = set(todo)
        while todo:
            edge = todo.pop()
            process(edge.start)
            process(edge.end)
        if len(network) > 1:  # a network requires two or more edges
            return network
        return set()

    def find_all_networks(self) -> Sequence[set[Edge]]:
        """Returns all separated networks in this deposit in ascending order of edge
        count.

        """
        edges = set(self.edges)
        networks: list[set[Edge]] = []
        while edges:
            edge = edges.pop()
            network = self.find_network(edge)
            if len(network):
                networks.append(network)
                edges -= network
            else:  # solitary edge
                edges.discard(edge)

        networks.sort(key=lambda n: len(n))
        return networks

    def find_leafs(self) -> Iterator[Edge]:
        """Yields all edges that have at least one end point without connection to other
        edges.
        """
        for edge in self.edges:
            if len(self.edges_linked_to(edge.start)) == 1:
                yield edge
            elif len(self.edges_linked_to(edge.end)) == 1:
                yield edge


def is_forward_connected(a: Edge, b: Edge, *, gap_tol=GAP_TOL) -> bool:
    """Returns ``True`` if the edges have a forward connection.

    Forward connection: distance from :attr:`a.end` to :attr:`b.start` <= gap_tol

    Args:
        a: first edge
        b: second edge
        gap_tol: maximum vertex distance to consider two edges as connected
    """
    return isclose(a.end, b.start, gap_tol=gap_tol)


def is_chain(edges: Sequence[Edge], *, gap_tol=GAP_TOL) -> bool:
    """Returns ``True`` if all edges in the sequence have a forward connection.

    Args:
        edges: sequence of edges
        gap_tol: maximum vertex distance to consider two edges as connected
    """
    return all(
        is_forward_connected(a, b, gap_tol=gap_tol) for a, b in zip(edges, edges[1:])
    )


def is_loop(edges: Sequence[Edge], *, gap_tol=GAP_TOL) -> bool:
    """Return ``True`` if the sequence of edges is a closed loop.

    Args:
        edges: sequence of edges
        gap_tol: maximum vertex distance to consider two edges as connected
    """
    if not is_chain(edges, gap_tol=gap_tol):
        return False
    return isclose(edges[-1].end, edges[0].start, gap_tol=gap_tol)


def is_loop_fast(edges: Sequence[Edge], *, gap_tol=GAP_TOL) -> bool:
    """Internal fast loop check."""
    return isclose(edges[-1].end, edges[0].start, gap_tol=gap_tol)


def length(edges: Sequence[Edge]) -> float:
    """Returns the length of a sequence of edges."""
    return sum(e.length for e in edges)


def shortest_chain(chains: Iterable[Sequence[Edge]]) -> Sequence[Edge]:
    """Returns the shortest chain of connected edges.

    .. note::

        This function does not verify if the input sequences are connected edges!

    """
    sorted_chains = sorted(chains, key=length)
    if sorted_chains:
        return sorted_chains[0]
    return tuple()


def longest_chain(chains: Iterable[Sequence[Edge]]) -> Sequence[Edge]:
    """Returns the longest chain of connected edges.

    .. Note::

        This function does not verify if the input sequences are connected edges!

    """
    sorted_chains = sorted(chains, key=length)
    if sorted_chains:
        return sorted_chains[-1]
    return tuple()


def reverse_chain(chain: Sequence[Edge]) -> list[Edge]:
    """Returns the reversed chain.

    The sequence order of the edges will be reversed as well as the start- and end
    points of the edges.
    """
    edges = list(chain)
    edges.reverse()
    return [edge.reversed() for edge in edges]


def find_sequential_chain(edges: Sequence[Edge], *, gap_tol=GAP_TOL) -> Sequence[Edge]:
    """Returns a simple chain beginning at the first edge.

    The search stops at the first edge without a forward connection from the previous
    edge.  Edges will be reversed if required to create connection.

    Args:
        edges: edges to be examined
        gap_tol: maximum vertex distance to consider two edges as connected

    Raises:
        TypeError: invalid data in sequence `edges`
    """
    edges = type_check(edges)
    if len(edges) < 2:
        return edges
    chain = [edges[0]]
    for edge in edges[1:]:
        last = chain[-1]
        if is_forward_connected(last, edge, gap_tol=gap_tol):
            chain.append(edge)
            continue
        reversed_edge = edge.reversed()
        if is_forward_connected(last, reversed_edge, gap_tol=gap_tol):
            chain.append(reversed_edge)
            continue
        break
    return chain


def find_all_sequential_chains(
    edges: Sequence[Edge], *, gap_tol=GAP_TOL
) -> Iterator[Sequence[Edge]]:
    """Yields all simple chains from sequence `edges`.

    The search progresses strictly in order of the input sequence. The search starts a
    new chain at every edge without a forward connection from the previous edge.
    Edges will be reversed if required to create connection.
    Each chain has one or more edges.

    Args:
        edges: sequence of edges
        gap_tol: maximum vertex distance to consider two edges as connected

    Raises:
        TypeError: invalid data in sequence `edges`
    """
    while edges:
        chain = find_sequential_chain(edges, gap_tol=gap_tol)
        edges = edges[len(chain) :]
        yield chain


def find_loop(deposit: Deposit, *, timeout: float = TIMEOUT) -> Sequence[Edge]:
    """Returns the first closed loop in `deposit`.

    Returns only simple loops, where all vertices have a degree of 2 (only two adjacent
    edges).

    .. note::

        This is a recursive backtracking algorithm with time complexity of O(n!).

    Args:
        deposit (Deposit): edge deposit
        timeout (float): timeout in seconds

    Raises:
        TimeoutError: search process has timed out
    """
    chains = find_all_simple_chains(deposit)
    if not chains:
        return tuple()

    gap_tol = deposit.gap_tol
    packed_edges: list[Edge] = []
    for chain in chains:
        if len(chain) > 1:
            if is_loop_fast(chain, gap_tol=gap_tol):
                return chain
            packed_edges.append(_wrap_simple_chain(chain))
        else:
            packed_edges.append(chain[0])
    deposit = Deposit(packed_edges, gap_tol=gap_tol)
    if len(deposit.edges) < 2:
        return tuple()
    return tuple(flatten(_find_loop_in_deposit(deposit, timeout=timeout)))


def _find_loop_in_deposit(deposit: Deposit, *, timeout=TIMEOUT) -> Sequence[Edge]:
    if len(deposit.edges) < 2:
        return tuple()

    finder = LoopFinder(deposit, timeout=timeout)
    loop = finder.find_any_loop()
    if loop:
        return loop
    return tuple()


def find_all_loops(
    deposit: Deposit, *, timeout: float = TIMEOUT
) -> Sequence[Sequence[Edge]]:
    """Returns all closed loops from `deposit`.

    Returns only simple loops, where all vertices have a degree of 2 (only two adjacent
    edges).  The result does not include reversed solutions.

    .. note::

        This is a recursive backtracking algorithm with time complexity of O(n!).

    Args:
        deposit (Deposit): edge deposit
        timeout (float): timeout in seconds

    Raises:
        TimeoutError: search process has timed out
    """
    chains = find_all_simple_chains(deposit)
    if not chains:
        return tuple()

    gap_tol = deposit.gap_tol
    solutions: list[Sequence[Edge]] = []
    packed_edges: list[Edge] = []
    for chain in chains:
        if len(chain) > 1:
            if is_loop_fast(chain, gap_tol=gap_tol):
                # these loops have no ambiguities (junctions)
                solutions.append(chain)
            else:
                packed_edges.append(_wrap_simple_chain(chain))
        else:
            packed_edges.append(chain[0])

    if not packed_edges:
        return solutions

    deposit = Deposit(packed_edges, gap_tol=gap_tol)
    if len(deposit.edges) < 2:
        return tuple()
    try:
        result = _find_all_loops_in_deposit(deposit, timeout=timeout)
    except TimeoutError as err:
        if err.solutions:
            solutions.extend(err.solutions)
            err.solutions = solutions
        raise
    solutions.extend(result)
    return _unwrap_simple_chains(solutions)


def _find_all_loops_in_deposit(
    deposit: Deposit, timeout=TIMEOUT
) -> Sequence[Sequence[Edge]]:
    solutions: list[Sequence[Edge]] = []
    finder = LoopFinder(deposit, timeout=timeout)
    for edge in deposit.edges:
        finder.search(edge)
    solutions.extend(finder)
    return solutions


def unique_chains(chains: Sequence[Sequence[Edge]]) -> Iterator[Sequence[Edge]]:
    """Filter duplicate chains and yields only unique chains.

    Yields the first chain for chains which have the same set of edges. The order of the
    edges is not important.
    """
    seen: set[frozenset[int]] = set()
    for chain in chains:
        key = frozenset(edge.id for edge in chain)
        if key not in seen:
            yield chain
            seen.add(key)


def type_check(edges: Sequence[Edge]) -> Sequence[Edge]:
    for edge in edges:
        if not isinstance(edge, Edge):
            raise TypeError(f"expected type <Edge>, got {str(type(edge))}")
    return edges


class _Vertex(Vec3):
    __slots__ = ("edge",)
    # for unknown reasons super().__init__(location) doesn't work, therefor no
    # _Vertex.__init__(self, location: Vec3, edge: Edge) constructor
    edge: Edge


def make_edge_vertex(location: Vec3, edge: Edge) -> _Vertex:
    vertex = _Vertex(location)
    vertex.edge = edge
    return vertex


class _SpatialSearchIndex:
    """Spatial search index of all edge vertices.

    (internal class)
    """

    def __init__(self, edges: Sequence[Edge]) -> None:
        vertices: list[_Vertex] = []
        for edge in edges:
            vertices.append(make_edge_vertex(edge.start, edge))
            vertices.append(make_edge_vertex(edge.end, edge))
        self._search_tree = rtree.RTree(vertices)

    @property
    def rtree(self) -> rtree.RTree[Vec3]:
        return self._search_tree

    def vertices_in_sphere(self, center: Vec3, radius: float) -> Sequence[_Vertex]:
        """Returns all vertices located around `center` with a max. distance of `radius`."""
        return tuple(self._search_tree.points_in_sphere(center, radius))

    def nearest_vertex(self, location: Vec3) -> _Vertex:
        """Returns the nearest vertex to the given location."""
        vertex, _ = self._search_tree.nearest_neighbor(location)
        return vertex


SearchSolutions: TypeAlias = Dict[Tuple[int, ...], Sequence[Edge]]


class LoopFinder:
    """Find closed loops in a :class:`Deposit` by a recursive backtracking algorithm.

    Finds only simple loops, where all vertices have only two adjacent edges.

    (internal class)
    """

    def __init__(self, deposit: Deposit, *, timeout=TIMEOUT) -> None:
        if len(deposit.edges) < 2:
            raise ValueError("two or more edges required")
        self._deposit = deposit
        self._timeout = timeout
        self._solutions: SearchSolutions = {}

    @property
    def gap_tol(self) -> float:
        return self._deposit.gap_tol

    def __iter__(self) -> Iterator[Sequence[Edge]]:
        return iter(self._solutions.values())

    def __len__(self) -> int:
        return len(self._solutions)

    def find_any_loop(self, start: Edge | None = None) -> Sequence[Edge]:
        """Returns the first loop found beginning with the given start edge or an
        arbitrary edge if `start` is None.
        """
        if start is None:
            start = self._deposit.edges[0]

        self.search(start, stop_at_first_loop=True)
        try:
            return next(iter(self._solutions.values()))
        except StopIteration:
            return tuple()

    def search(self, start: Edge, stop_at_first_loop: bool = False) -> None:
        """Searches for all loops that begin at the given start edge and contain
        only vertices of degree 2.

        These are not all possible loops in the edge deposit!

        Raises:
            TimeoutError: search process has timed out, intermediate results are attached
                TimeoutError.data

        """
        deposit = self._deposit
        gap_tol = self.gap_tol
        start_point = start.start
        watchdog = Watchdog(self._timeout)
        todo: list[tuple[Edge, ...]] = [(start,)]  # "unlimited" recursion stack
        while todo:
            if watchdog.has_timed_out:
                raise TimeoutError(
                    "search process has timed out",
                    solutions=tuple(self._solutions.values()),  # noqa
                )
            chain = todo.pop()
            last_edge = chain[-1]
            end_point = last_edge.end
            candidates = deposit.edges_linked_to(end_point, radius=gap_tol)
            # edges must be unique in a loop
            survivors = set(candidates) - set(chain)
            for edge in survivors:
                if isclose(end_point, edge.start, gap_tol=gap_tol):
                    next_edge = edge
                else:
                    next_edge = edge.reversed()
                last_point = next_edge.end
                if isclose(last_point, start_point, gap_tol=gap_tol):
                    self.add_solution(chain + (next_edge,))
                    if stop_at_first_loop:
                        return
                # Add only chains to the stack that have vertices of max degree 2.
                # If the new end point is in the chain, a vertex of degree 3 would be
                # created. (loop check is done)
                elif not any(
                    isclose(last_point, e.end, gap_tol=gap_tol) for e in chain
                ):
                    todo.append(chain + (next_edge,))

    def add_solution(self, loop: Sequence[Edge]) -> None:
        solutions = self._solutions
        key = loop_key(loop)
        if key in solutions or loop_key(loop, reverse=True) in solutions:
            return
        solutions[key] = loop


def loop_key(edges: Sequence[Edge], *, reverse=False) -> tuple[int, ...]:
    """Returns a normalized key.

    The key is rotated to begin with the smallest edge id.
    """
    if reverse:
        ids = tuple(edge.id for edge in reversed(edges))
    else:
        ids = tuple(edge.id for edge in edges)
    index = ids.index(min(ids))
    if index:
        ids = ids[index:] + ids[:index]
    return ids


def find_all_simple_chains(deposit: Deposit) -> Sequence[Sequence[Edge]]:
    """Returns all simple chains from `deposit`.

    Each chains starts and ends at a leaf (degree of 1) or a junction (degree greater 2).
    All vertices between the start- and end vertex have a degree of 2.
    The result doesn't include reversed solutions.
    """
    if len(deposit.edges) < 1:
        return tuple()
    solutions: list[Sequence[Edge]] = []
    edges = set(deposit.edges)
    while edges:
        chain = find_simple_chain(deposit, edges.pop())
        solutions.append(chain)
        edges -= set(chain)
    return solutions


def find_simple_chain(deposit: Deposit, start: Edge) -> Sequence[Edge]:
    """Returns a simple chain containing `start` edge.

    A simple chain start and ends at a leaf or a junction.

    All connected edges have vertices of degree 2, except the first and last vertex.
    The first and the last vertex have a degree of 1 (leaf) or greater 2 (junction).
    """
    forward_chain = _simple_forward_chain(deposit, start)
    if is_loop_fast(forward_chain, gap_tol=deposit.gap_tol):
        return forward_chain
    backwards_chain = _simple_forward_chain(deposit, start.reversed())
    if len(backwards_chain) == 1:
        return forward_chain
    backwards_chain.reverse()
    backwards_chain.pop()  # reversed start
    return [edge.reversed() for edge in backwards_chain] + forward_chain


def _simple_forward_chain(deposit: Deposit, edge: Edge) -> list[Edge]:
    """Returns a simple chain beginning with `edge`.

    All connected edges have vertices of degree 2, expect for the last edge.
    The last edge has an end-vertex of degree 1 (leaf) or greater 2 (junction).
    """
    gap_tol = deposit.gap_tol
    chain = [edge]
    while True:
        last = chain[-1]
        linked = deposit.edges_linked_to(last.end, gap_tol)
        if len(linked) != 2:  # no junctions allowed!
            return chain
        if linked[0] == last:
            edge = linked[1]
        else:
            edge = linked[0]
        if isclose(last.end, edge.start, gap_tol=gap_tol):
            chain.append(edge)
        else:
            chain.append(edge.reversed())
        if is_loop_fast(chain, gap_tol=gap_tol):
            return chain


def is_wrapped_chain(edge: Edge) -> bool:
    """Returns ``True`` if `edge` is a wrapper for linked edges."""
    return isinstance(edge.payload, EdgeWrapper)


def wrap_simple_chain(chain: Sequence[Edge], *, gap_tol=GAP_TOL) -> Edge:
    """Wraps a sequence of linked edges (simple chain) into a single edge.

    Two or more linked edges required. Closed loops cannot be wrapped into a single
    edge.

    Raises:
        ValueError: less than two edges; not a chain; chain is a closed loop

    """
    if len(chain) < 2:
        raise ValueError("two or more linked edges required")
    if is_chain(chain, gap_tol=gap_tol):
        if is_loop_fast(chain, gap_tol=gap_tol):
            raise ValueError("closed loop cannot be wrapped into a single edge")
        return _wrap_simple_chain(chain)
    raise ValueError("edges are not connected")


def unwrap_simple_chain(edge: Edge) -> Sequence[Edge]:
    """Unwraps a simple chain which is wrapped into a single edge."""
    if isinstance(edge.payload, EdgeWrapper):
        return _unwrap_simple_chain(edge)
    return (edge,)


class EdgeWrapper:
    """Internal class to wrap a sequence of linked edges."""

    __slots__ = ("edges",)

    def __init__(self, edges: Sequence[Edge]) -> None:
        self.edges: Sequence[Edge] = tuple(edges)


def _wrap_simple_chain(edges: Sequence[Edge]) -> Edge:
    return make_edge(
        edges[0].start, edges[-1].end, length(edges), payload=EdgeWrapper(edges)
    )


def _unwrap_simple_chain(edge: Edge) -> Sequence[Edge]:
    wrapper = edge.payload
    assert isinstance(wrapper, EdgeWrapper)
    if edge.is_reverse:
        return tuple(e.reversed() for e in reversed(wrapper.edges))
    else:
        return wrapper.edges


def _unwrap_simple_chains(chains: Iterable[Iterable[Edge]]) -> Sequence[Sequence[Edge]]:
    return tuple(tuple(flatten(chain)) for chain in chains)


def flatten(edges: Edge | Iterable[Edge]) -> Iterator[Edge]:
    """Yields all edges from any nested structure of edges as a flat stream of edges."""
    edge: Edge
    if not isinstance(edges, Edge):
        for edge in edges:
            yield from flatten(edge)
    else:
        edge = edges
        if is_wrapped_chain(edge):
            yield from flatten(_unwrap_simple_chain(edge))
        else:
            yield edge


def chain_key(edges: Sequence[Edge], reverse=False) -> tuple[int, ...]:
    """Returns a normalized key."""
    if reverse:
        return tuple(edge.id for edge in reversed(edges))
    else:
        return tuple(edge.id for edge in edges)


def find_all_open_chains(
    deposit: Deposit, timeout: float = TIMEOUT
) -> Sequence[Sequence[Edge]]:
    """Returns all open chains from `deposit`.

    Returns only simple chains ending on both sides with a leaf.
    A leaf is a vertex of degree 1 without further connections.
    All vertices have a degree of 2 except for the leafs at the start and end.
    The result does not include reversed solutions or closed loops.

    .. note::

        This is a recursive backtracking algorithm with time complexity of O(n!).

    Args:
        deposit (Deposit): edge deposit
        timeout (float): timeout in seconds

    Raises:
        TimeoutError: search process has timed out
    """

    finder = OpenChainFinder(deposit, timeout)
    for edge in deposit.find_leafs():
        finder.search(edge)
    solutions = finder.solutions
    solutions.sort(key=lambda x: len(x))
    return solutions


class OpenChainFinder:
    def __init__(self, deposit: Deposit, timeout=TIMEOUT):
        self.deposit = deposit
        self.solution_keys: set[tuple[int, ...]] = set()
        self.solutions: list[Sequence[Edge]] = []
        self.watchdog = Watchdog(timeout)

    def search(self, edge: Edge) -> None:
        forward_chains = self.forward_search(edge)
        self.reverse_search(forward_chains)

    def forward_search(self, edge: Edge) -> list[tuple[Edge, ...]]:
        deposit = self.deposit
        gap_tol = deposit.gap_tol
        watchdog = self.watchdog

        forward_chains: list[tuple[Edge, ...]] = []
        todo: list[tuple[Edge, ...]] = [(edge,)]
        while todo:
            if watchdog.has_timed_out:
                raise TimeoutError("search has timed out")
            chain = todo.pop()
            start_point = chain[-1].end
            candidates = deposit.edges_linked_to(start_point)
            backwards_edges = set(candidates) - set(chain)
            if backwards_edges:
                for edge in backwards_edges:
                    if not isclose(start_point, edge.start, gap_tol=gap_tol):
                        edge = edge.reversed()
                    # Add only chains to the stack that have vertices of max degree 2.
                    # If the new end point is in the chain, a vertex of degree 3 would be
                    # created. (loop check is done)
                    last_point = edge.end
                    if not any(
                        isclose(last_point, e.end, gap_tol=gap_tol) for e in chain
                    ):
                        todo.append(chain + (edge,))
            else:
                forward_chains.append(chain)
        return forward_chains

    def reverse_search(self, forward_chains: list[tuple[Edge, ...]]) -> None:
        deposit = self.deposit
        gap_tol = deposit.gap_tol
        watchdog = self.watchdog
        todo = forward_chains
        while todo:
            if watchdog.has_timed_out:
                raise TimeoutError("search has timed out", solutions=self.solutions)
            chain = todo.pop()
            start_point = chain[0].start
            candidates = deposit.edges_linked_to(start_point)
            backwards_edges = set(candidates) - set(chain)
            if backwards_edges:
                for edge in backwards_edges:
                    if not isclose(start_point, edge.end, gap_tol=gap_tol):
                        edge = edge.reversed()
                    # Add only chains to the stack that have vertices of max degree 2.
                    # If the new end point is in the chain, a vertex of degree 3 would be
                    # created.
                    new_start_point = edge.start
                    if not any(
                        isclose(new_start_point, e.end, gap_tol=gap_tol) for e in chain
                    ):
                        todo.append((edge,) + chain)
            else:
                self.add_solution(chain)

    def add_solution(self, solution: Sequence[Edge]) -> None:
        keys = self.solution_keys
        key = chain_key(solution)
        if key in keys:
            return
        keys.add(key)
        key = chain_key(solution, reverse=True)
        if key in keys:
            return
        keys.add(key)
        self.solutions.append(solution)


def count_checker(count: int):
    """Returns a function that checks if a given sequence of edges has at least `count`
    vertices.
    """

    def has_min_edge_count(edges: Sequence[Edge]) -> bool:
        return len(edges) >= count

    return has_min_edge_count


def line_checker(abs_tol=ABS_TOL):
    """Returns a function that checks if two input edges are congruent."""

    def is_congruent_line(a: Edge, b: Edge) -> bool:
        return (
            a.start.isclose(b.start, abs_tol=abs_tol)
            and a.end.isclose(b.end, abs_tol=abs_tol)
        ) or (
            a.start.isclose(b.end, abs_tol=abs_tol)
            and a.end.isclose(b.start, abs_tol=abs_tol)
        )

    return is_congruent_line


def filter_coincident_edges(
    deposit: Deposit, eq_fn: Callable[[Edge, Edge], bool] = line_checker()
) -> Sequence[Edge]:
    """Returns all edges from deposit that are not coincident to any other edge in the
    deposit. Coincident edges are detected by the given `eq_fn` function.
    """
    edges = set(deposit.edges)
    unique_edges: list[Edge] = []
    while edges:
        edge = edges.pop()
        unique_edges.append(edge)
        candidates = set(deposit.edges_linked_to(edge.start))
        candidates.update(deposit.edges_linked_to(edge.end))
        candidates.discard(edge)
        for candidate in candidates:
            if eq_fn(edge, candidate):
                edges.discard(candidate)
    return unique_edges


def filter_close_vertices(rt: rtree.RTree[Vec3], *, gap_tol: float) -> set[Vec3]:
    """Returns all vertices from a :class:`RTree` and filters vertices that are closer
    than radius `gap_tol` to another vertex.

    Vertices that are close to another vertex are filtered out, so none of the returned
    vertices has another vertex within the range of `gap_tol`.  It is not determined
    which of close vertices is returned.
    """
    # RTree cannot be empty!
    todo = set(rt)
    merged: set[Vec3] = set([todo.pop()])
    for vertex in todo:
        for candidate in rt.points_in_sphere(vertex, gap_tol):
            if candidate in merged:
                continue
            if not any(isclose(candidate, v, gap_tol=gap_tol) for v in merged):
                merged.add(candidate)
    return merged


def sort_edges_to_base(
    edges: Iterable[Edge], base: Edge, *, gap_tol=GAP_TOL
) -> list[Edge]:
    """Returns a list of `edges` sorted in counter-clockwise order in relation to the
    `base` edge.

    The `base` edge represents zero radians.
    All edges have to be connected to end-vertex of the `base` edge.
    This is a pure 2D algorithm and ignores all z-axis values.

    Args:
        edges: list of edges to sort
        base: base edge for sorting, represents 0-radian
        gap_tol: maximum vertex distance to consider two edges as connected

    Raises:
        ValueError: edge is not connected to center

    """

    def angle(edge: Edge) -> float:
        s = Vec2(edge.start)
        e = Vec2(edge.end)
        if connection_point.distance(s) <= gap_tol:
            edge_angle = (e - s).angle
        elif connection_point.distance(e) <= gap_tol:
            edge_angle = (s - e).angle
        else:
            raise ValueError(f"edge {edge!s} not connected to center")
        return (edge_angle - base_angle) % math.tau

    connection_point = Vec2(base.end)
    base_angle = (Vec2(base.start) - connection_point).angle
    return sorted(edges, key=angle)


def find_loop_by_edge(deposit: Deposit, start: Edge, clockwise=True) -> Sequence[Edge]:
    """Returns the first loop found including the given `start` edge.

    Returns an empty sequence if no loop was found.

    Args:
        deposit (Deposit): edge deposit
        start (Edge): starting edge of the search
        clockwise (bool): search orientation, counter-clockwise when ``False``

    """
    if len(deposit.edges) < 2:
        return tuple()
    gap_tol = deposit.gap_tol
    chain = [start]
    chain_set = set(chain)
    start_point = start.start

    while True:
        last_edge = chain[-1]
        end_point = last_edge.end
        candidates = deposit.edges_linked_to(end_point, radius=gap_tol)
        # edges must be unique in a loop
        survivors = set(candidates) - chain_set
        count = len(survivors)
        if count == 0:
            return tuple()  # dead end
        if count > 1:
            sorted_edges = sort_edges_to_base(survivors, last_edge, gap_tol=gap_tol)
            if clockwise:
                next_edge = sorted_edges[-1]  # first clockwise edge
            else:
                next_edge = sorted_edges[0]  # first counter-clockwise edge
        else:
            next_edge = survivors.pop()
        if not isclose(next_edge.start, end_point, gap_tol=gap_tol):
            next_edge = next_edge.reversed()
        chain.append(next_edge)
        if isclose(next_edge.end, start_point, gap_tol=gap_tol):
            return chain  # found a closed loop
        chain_set.add(next_edge)


def subtract_edges(base: Iterable[Edge], edges: Iterable[Edge]) -> list[Edge]:
    """Returns all edges from the iterable `base` that do not exist in the iterable
    `edges` e.g. remove solutions from the search space.
    """
    edge_ids = set(edge.id for edge in edges)
    return [edge for edge in base if edge.id not in edge_ids]



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.math import Vec2
from ezdxf.addons.drawing.backend import BackendProperties, BkPoints2d, BkPath2d
from ezdxf.addons.drawing import json
from ezdxf import path


class TestCustomJSONBackend:
    @pytest.fixture
    def backend(self):
        return json.CustomJSONBackend()

    @pytest.fixture
    def properties(self):
        return BackendProperties(color="#ff0000", lineweight=0.25, layer="One")

    def test_draw_point(
        self, backend: json.CustomJSONBackend, properties: BackendProperties
    ):
        backend.draw_point(Vec2(13, 17), properties)
        result = backend.get_json_data()
        point = result[0]
        assert point["type"] == "point"
        assert point["properties"]["color"] == "#ff0000"
        assert point["properties"]["stroke-width"] == 0.25
        assert point["properties"]["layer"] == "One"
        assert point["geometry"] == [13, 17]

    def test_draw_solid_lines(
        self, backend: json.CustomJSONBackend, properties: BackendProperties
    ):
        backend.draw_solid_lines(
            [
                (Vec2(0, 0), Vec2(1, 0)),
                (Vec2(2, 0), Vec2(3, 0)),
                (Vec2(4, 0), Vec2(5, 0)),
            ],
            properties,
        )
        result = backend.get_json_data()
        lines = result[0]
        assert lines["type"] == "lines"
        assert lines["geometry"] == [(0, 0, 1, 0), (2, 0, 3, 0), (4, 0, 5, 0)]

    def test_draw_path(
        self, backend: json.CustomJSONBackend, properties: BackendProperties
    ):
        p = path.Path((2, 2))
        p.line_to((4, 2))
        p.curve3_to((8, 2), (6, 3))
        p.curve4_to((11, 2), (9, 3), (10, 3))

        backend.draw_path(BkPath2d(p), properties)
        result = backend.get_json_data()
        json_path = result[0]
        assert json_path["type"] == "path"
        assert json_path["geometry"] == [
            ("M", 2, 2),
            ("L", 4, 2),
            ("Q", 6, 3, 8, 2),
            ("C", 9, 3, 10, 3, 11, 2),
        ]

    def test_draw_filled_polygon(
        self, backend: json.CustomJSONBackend, properties: BackendProperties
    ):  
        polygon = BkPoints2d(Vec2.list([(0, 0), (2, 0), (1, 1)]))
        backend.draw_filled_polygon(polygon, properties)
        result = backend.get_json_data()
        json_polygon = result[0]
        assert json_polygon["type"] == "filled-polygon"
        assert json_polygon["geometry"] == [(0, 0), (2, 0), (1, 1), (0, 0)]

    def test_json_str(
        self, backend: json.CustomJSONBackend, properties: BackendProperties
    ):
        backend.draw_point(Vec2(13, 17), properties)
        result = backend.get_string(indent=4)
        assert result == """[
    {
        "type": "point",
        "properties": {
            "color": "#ff0000",
            "stroke-width": 0.25,
            "layer": "One"
        },
        "geometry": [
            13.0,
            17.0
        ]
    }
]"""

if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
import math
import pytest
import ezdxf
from ezdxf.math import UCS, Vec3
from ezdxf.layouts import Modelspace


@pytest.fixture(scope="module")
def msp() -> Modelspace:
    doc = ezdxf.new(setup=True)
    msp = doc.modelspace()
    base = Vec3(0, 5)
    p1 = Vec3(-4, 3)
    p2 = Vec3(-1, 0)
    p4 = Vec3(4, 3)
    p3 = Vec3(1, 0)
    ucs = UCS(origin=(5, 7, -3)).rotate_local_x(math.radians(45))
    dim = msp.add_angular_dim_2l(
        base=base, line1=(p1, p2), line2=(p3, p4), dimstyle="EZ_CURVED"
    )
    dim.render(ucs=ucs)
    dim.dimension.explode()
    return msp


# Run example "examples/render/explode_3d_dimension.py" to create the DXF files
# for a visual comparison.

EXPECTED_EXT = Vec3(0.0, -0.7071067811865476, 0.7071067811865476)


def test_exploded_arrow_blocks(msp: Modelspace):
    b1, b2 = msp.query("INSERT")
    assert EXPECTED_EXT.isclose(b1.dxf.extrusion)
    assert EXPECTED_EXT.isclose(b2.dxf.extrusion)
    assert b1.dxf.insert.isclose(
        (9.242640687119284, 6.0710678118654755, -7.071067811865476)
    )
    assert b2.dxf.insert.isclose(
        (0.7573593128807152, 6.0710678118654755, -7.071067811865476)
    )


def test_exploded_arc(msp: Modelspace):
    arc = msp.query("ARC")[0]
    assert EXPECTED_EXT.isclose(arc.dxf.extrusion)
    assert arc.dxf.center.isclose((5.0, 1.8284271247461898, -7.071067811865476))
    assert math.isclose(arc.dxf.radius, 6.0)
    assert math.isclose(arc.dxf.start_angle, 47.38732414637843)
    assert math.isclose(arc.dxf.end_angle, 132.61267585362157)


def test_exploded_mtext(msp: Modelspace):
    mtext = msp.query("MTEXT")[0]
    assert EXPECTED_EXT.isclose(mtext.dxf.extrusion)
    assert mtext.dxf.insert.isclose(
        (5.0, 10.694632931699713, 0.6946329316997115)
    )
    assert mtext.dxf.text_direction.isclose((1.0, 0, 0))


def test_exploded_extension_line_1(msp: Modelspace):
    ext1, _ = msp.query("LINE")
    assert ext1.dxf.start.isclose(
        (9.08838834764832, 9.183820343559644, -0.8161796564403581)
    )
    assert ext1.dxf.end.isclose(
        (9.507805730064241, 9.480393218813454, -0.5196067811865479)
    )


def test_exploded_extension_line_2(msp: Modelspace):
    _, ext2 = msp.query("LINE")
    assert ext2.dxf.start.isclose(
        (0.9116116523516817, 9.183820343559644, -0.8161796564403581)
    )
    assert ext2.dxf.end.isclose(
        (0.49219426993575954, 9.480393218813454, -0.5196067811865479)
    )


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
from typing import Iterable, Sequence, List
import pytest
import math
from functools import partial
from ezdxf.math import Vec2, BoundingBox2d, area, UVec
from ezdxf.render import forms
from ezdxf.math._mapbox_earcut import earcut as _py_earcut

CYTHON = "Cython"
try:
    from ezdxf.acc.mapbox_earcut import earcut as _cy_earcut
except ImportError:
    CYTHON = "CPython"
    _cy_earcut = _py_earcut


def earcut_driver(
    exterior: Iterable[UVec],
    holes: Iterable[Iterable[UVec]] = None,
    func=_py_earcut,
) -> List[Sequence[Vec2]]:
    points: List[Vec2] = Vec2.list(exterior)
    if len(points) == 0:
        return []
    holes_: List[List[Vec2]] = []
    if holes:
        holes_ = [Vec2.list(hole) for hole in holes]
    return func(points, holes_)


py_earcut = partial(earcut_driver, func=_py_earcut)
cy_earcut = partial(earcut_driver, func=_cy_earcut)
functions = [py_earcut, cy_earcut]
names = ("CPython", CYTHON)


def total_area(triangles: Iterable[Sequence[Vec2]]):
    area = 0.0
    sides = []
    for triangle in triangles:
        sides.clear()
        for i in range(3):
            pt = triangle[i]
            pt2 = triangle[(i + 1) % 3]
            sides.append(pt.distance(pt2))
        a, b, c = sides
        area += 0.25 * math.sqrt(
            (a + b + c) * (-a + b + c) * (a - b + c) * (a + b - c)
        )
    return area


@pytest.mark.parametrize("earcut", functions, ids=names)
def test_triangulate_ccw_square(earcut):
    square = forms.square(2)
    triangles = earcut(square)
    assert len(triangles) == 2
    assert total_area(triangles) == pytest.approx(4.0)


@pytest.mark.parametrize("earcut", functions, ids=names)
def test_triangulate_cw_square(earcut):
    square = list(reversed(forms.square(2)))
    triangles = earcut(square)
    assert len(triangles) == 2
    assert total_area(triangles) == pytest.approx(4.0)


@pytest.mark.parametrize("earcut", functions, ids=names)
def test_triangulate_concave_gear_shape(earcut):
    square = list(
        forms.gear(32, top_width=1, bottom_width=3, height=2, outside_radius=10)
    )
    triangles = earcut(square)
    assert len(triangles) == 126
    assert total_area(triangles) == pytest.approx(265.17899685816224)


@pytest.mark.parametrize("earcut", functions, ids=names)
def test_triangulate_square_with_square_hole(earcut):
    square = forms.square(4, center=True)
    hole = forms.square(2, center=True)
    triangles = earcut(square, holes=[hole])
    assert len(triangles) == 8
    assert total_area(triangles) == pytest.approx(16.0 - 4.0)


@pytest.mark.parametrize("earcut", functions, ids=names)
def test_triangulate_square_with_two_holes(earcut):
    square = list(forms.square(4, center=True))
    hole0 = list(forms.translate(forms.square(1, center=True), (-1, -1)))
    hole1 = list(forms.translate(forms.square(1, center=True), (1, 1)))
    holes = [hole0, hole1]
    triangles = earcut(square, holes=holes)
    assert len(triangles) == 14
    assert total_area(triangles) == pytest.approx(16.0 - 2.0)


@pytest.mark.parametrize("earcut", functions, ids=names)
def test_triangulate_square_with_steiner_point(earcut):
    square = forms.square(4, center=True)
    steiner_point = [(0, 0)]  # defined as a hole with a single point
    triangles = earcut(square, holes=[steiner_point])
    assert len(triangles) == 4
    assert total_area(triangles) == pytest.approx(16.0)


@pytest.mark.parametrize("earcut", functions, ids=names)
def test_empty_exterior(earcut):
    triangles = list(earcut([]))
    assert len(triangles) == 0


@pytest.mark.parametrize("earcut", functions, ids=names)
def test_empty_holes(earcut):
    square = forms.square(2)
    assert len(earcut(square, [[]])) == 2


@pytest.mark.parametrize("earcut", functions, ids=names)
def test_polygon_data0(polygon_data0, earcut):
    data = polygon_data0
    triangles = list(earcut(data.vertices))
    area0 = area(data.vertices)
    area1 = total_area(triangles)
    absolute_error = abs(area0 - area1)

    assert math.isclose(
        area0, area1
    ), "{}: area absolute error ({:.3f} - {:.3f} = {:.6f})".format(
        data.name,
        area0,
        area1,
        absolute_error,
    )
    bbox0 = BoundingBox2d(data.vertices)
    bbox1 = BoundingBox2d()
    for t in triangles:
        bbox1.extend(t)
    assert bbox0.extmin.isclose(bbox1.extmin)
    assert bbox0.extmax.isclose(bbox1.extmax)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2023, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING
from ezdxf.fonts import fonts
from ezdxf.fonts.font_measurements import FontMeasurements

from .text_renderer import TextRenderer

if TYPE_CHECKING:
    from ezdxf.npshapes import NumpyPath2d


class UnifiedTextRenderer(TextRenderer):
    """This text renderer supports .ttf, .ttc, .otf, .shx, .shp and .lff fonts.

    The resolving order for .shx fonts is applied in the RenderContext.add_text_style()
    method.

    """

    def __init__(self) -> None:
        self._font_cache: dict[str, fonts.AbstractFont] = dict()

    def get_font(self, font_face: fonts.FontFace) -> fonts.AbstractFont:
        if not font_face.filename and font_face.family:
            found = fonts.find_best_match(
                family=font_face.family,
                weight=700 if font_face.is_bold else 400,
                italic=font_face.is_italic,
            )
            if found is not None:
                font_face = found
        key = font_face.filename.lower()
        try:
            return self._font_cache[key]
        except KeyError:
            pass
        abstract_font = fonts.make_font(font_face.filename, 1.0)
        self._font_cache[key] = abstract_font
        return abstract_font

    def is_stroke_font(self, font_face: fonts.FontFace) -> bool:
        abstract_font = self.get_font(font_face)
        return abstract_font.font_render_type == fonts.FontRenderType.STROKE

    def get_font_measurements(
        self, font_face: fonts.FontFace, cap_height: float = 1.0
    ) -> FontMeasurements:
        abstract_font = self.get_font(font_face)
        return abstract_font.measurements.scale(cap_height)

    def get_text_path(
        self, text: str, font_face: fonts.FontFace, cap_height: float = 1.0
    ) -> NumpyPath2d:
        abstract_font = self.get_font(font_face)
        return abstract_font.text_path_ex(text, cap_height)

    def get_text_glyph_paths(
        self, text: str, font_face: fonts.FontFace, cap_height: float = 1.0
    ) -> list[NumpyPath2d]:
        abstract_font = self.get_font(font_face)
        return abstract_font.text_glyph_paths(text, cap_height)

    def get_text_line_width(
        self,
        text: str,
        font_face: fonts.FontFace,
        cap_height: float = 1.0,
    ) -> float:
        abstract_font = self.get_font(font_face)
        return abstract_font.text_width_ex(text, cap_height)



# Copyright (c) 2018-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
import logging
from typing import Iterable, TYPE_CHECKING, Optional
from collections import OrderedDict

from .const import DXFStructureError
from .tags import group_tags, DXFTag, Tags
from .extendedtags import ExtendedTags
from ezdxf.entities import factory

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.entities import DXFEntity
    from ezdxf.eztypes import SectionDict

logger = logging.getLogger("ezdxf")


def load_dxf_structure(
    tagger: Iterable[DXFTag], ignore_missing_eof: bool = False
) -> SectionDict:
    """Divide input tag stream from tagger into DXF structure entities.
    Each DXF structure entity starts with a DXF structure (0, ...) tag,
    and ends before the next DXF structure tag.

    Generated structure:

    each entity is a Tags() object

    {
        # 1. section, HEADER section consist only of one entity
        'HEADER': [entity],
        'CLASSES': [entity, entity, ...],  # 2. section
        'TABLES': [entity, entity, ...],   # 3. section
        ...
        'OBJECTS': [entity, entity, ...],
    }

    {
        # HEADER section consist only of one entity
        'HEADER': [(0, 'SECTION'), (2, 'HEADER'), .... ],
        'CLASSES': [
            [(0, 'SECTION'), (2, 'CLASSES')],
            [(0, 'CLASS'), ...],
            [(0, 'CLASS'), ...]
        ],
        'TABLES': [
            [(0, 'SECTION'), (2, 'TABLES')],
            [(0, 'TABLE'), (2, 'VPORT')],
            [(0, 'VPORT'), ...],
            ... ,
            [(0, 'ENDTAB')]
        ],
        ...
        'OBJECTS': [
            [(0, 'SECTION'), (2, 'OBJECTS')],
            ... ,
        ]
    }

    Args:
        tagger: generates DXFTag() entities from input data
        ignore_missing_eof: raises DXFStructureError() if False and EOF tag is
            not present, set to True only in tests

    Returns:
        dict of sections, each section is a list of DXF structure entities
        as Tags() objects

    """

    def inside_section() -> bool:
        if len(section):
            return section[0][0] == (0, "SECTION")  # first entity, first tag
        return False

    def outside_section() -> bool:
        if len(section):
            return section[0][0] != (0, "SECTION")  # first entity, first tag
        return True

    sections: SectionDict = OrderedDict()
    section: list[Tags] = []
    eof = False
    # The structure checking here should not be changed, ezdxf expect a valid
    # DXF file, to load messy DXF files exist an (future) add-on
    # called 'recover'.

    for entity in group_tags(tagger):
        tag = entity[0]
        if tag == (0, "SECTION"):
            if inside_section():
                raise DXFStructureError("DXFStructureError: missing ENDSEC tag.")
            if len(section):
                logger.warning(
                    "DXF Structure Warning: found tags outside a SECTION, "
                    "ignored by ezdxf."
                )
            section = [entity]
        elif tag == (0, "ENDSEC"):
            # ENDSEC tag is not collected.
            if outside_section():
                raise DXFStructureError(
                    "DXFStructureError: found ENDSEC tag without previous "
                    "SECTION tag."
                )
            section_header = section[0]

            if len(section_header) < 2 or section_header[1].code != 2:
                raise DXFStructureError(
                    "DXFStructureError: missing required section NAME tag "
                    "(2, name) at start of section."
                )
            name_tag = section_header[1]
            sections[name_tag.value] = section  # type: ignore
            # Collect tags outside of sections, but ignore it.
            section = []
        elif tag == (0, "EOF"):
            # EOF tag is not collected.
            if eof:
                logger.warning("DXF Structure Warning: found more than one EOF tags.")
            eof = True
        else:
            section.append(entity)
    if inside_section():
        raise DXFStructureError("DXFStructureError: missing ENDSEC tag.")
    if not eof and not ignore_missing_eof:
        raise DXFStructureError("DXFStructureError: missing EOF tag.")
    return sections


def load_dxf_entities(
    entities: Iterable[Tags], doc: Optional[Drawing] = None
) -> Iterable[DXFEntity]:
    for entity in entities:
        yield factory.load(ExtendedTags(entity), doc)


def load_and_bind_dxf_content(sections: dict, doc: Drawing) -> None:
    # HEADER has no database entries.
    db = doc.entitydb
    for name in ["TABLES", "CLASSES", "ENTITIES", "BLOCKS", "OBJECTS"]:
        if name in sections:
            section = sections[name]
            for index, entity in enumerate(load_dxf_entities(section, doc)):
                handle = entity.dxf.get("handle")
                if handle and handle in db:
                    logger.warning(
                        f"Found non-unique entity handle #{handle}, data validation is required."
                    )
                # Replace Tags() by DXFEntity() objects
                section[index] = entity
                # Bind entities to the DXF document:
                factory.bind(entity, doc)



# Copyright (c) 2013-2019, Manfred Moitzi
# License: MIT-License
import pytest

from ezdxf.queryparser import EntityQueryParser, ParseException, InfixBoolQuery


class TestEntityQueryParserWithoutAttributes:
    def test_without_wildcards(self):
        result = EntityQueryParser.parse_string("LINE", parse_all=True)
        name = result.EntityQuery[0]
        assert "LINE" == name

    def test_two_entity_names(self):
        result = EntityQueryParser.parse_string("LINE CIRCLE", parse_all=True)
        assert "LINE" == result.EntityQuery[0]
        assert "CIRCLE" == result.EntityQuery[1]

    def test_star_wildcard(self):
        result = EntityQueryParser.parse_string("*", parse_all=True)
        name = result.EntityQuery[0]
        assert "*" == name

    def test_wrong_star_wildcard(self):
        with pytest.raises(ParseException):
            EntityQueryParser.parse_string("LIN*", parse_all=True)

    def test_star_wildcard_2(self):
        result = EntityQueryParser.parse_string("* !LINE", parse_all=True)
        assert result.EntityQuery[0] == "*"
        assert result.EntityQuery[1] == "!LINE"

    def test_star_wildcard_3(self):
        with pytest.raises(ParseException):
            EntityQueryParser.parse_string("!LINE *", parse_all=True)

    def test_star_wildcard_4(self):
        with pytest.raises(ParseException):
            EntityQueryParser.parse_string("* LINE", parse_all=True)


class TestEntityQueryParserWithAttributes:
    def test_empty_attribute_list_not_allowed(self):
        with pytest.raises(ParseException):
            EntityQueryParser.parse_string("LINE[]", parse_all=True)

    def test_one_attribute(self):
        result = EntityQueryParser.parse_string(
            'LINE[layer=="0"]', parse_all=True
        )
        assert "LINE" == result.EntityQuery[0]
        assert ("layer", "==", "0") == tuple(result.AttribQuery)

    def test_double_quoted_attributes(self):
        result = EntityQueryParser.parse_string(
            'LINE[layer=="0"]', parse_all=True
        )
        assert "LINE" == result.EntityQuery[0]
        assert ("layer", "==", "0") == tuple(result.AttribQuery)

    def test_single_quoted_attributes(self):
        result = EntityQueryParser.parse_string(
            "LINE[layer=='0']", parse_all=True
        )
        assert "LINE" == result.EntityQuery[0]
        assert ("layer", "==", "0") == tuple(result.AttribQuery)

    def test_attribute_name_with_underscore(self):
        result = EntityQueryParser.parse_string(
            "HATCH[solid_fill==0]", parse_all=True
        )
        assert "HATCH" == result.EntityQuery[0]
        assert ("solid_fill", "==", 0) == tuple(result.AttribQuery)

    def test_star_with_one_attribute(self):
        result = EntityQueryParser.parse_string('*[layer=="0"]', parse_all=True)
        assert "*" == result.EntityQuery[0]
        assert 3 == len(result.AttribQuery)
        assert ("layer", "==", "0") == tuple(result.AttribQuery)

    def test_relation_lt(self):
        result = EntityQueryParser.parse_string('*[layer<"0"]', parse_all=True)
        assert ("layer", "<", "0") == tuple(result.AttribQuery)

    def test_relation_le(self):
        result = EntityQueryParser.parse_string('*[layer<="0"]', parse_all=True)
        assert ("layer", "<=", "0") == tuple(result.AttribQuery)

    def test_relation_eq(self):
        result = EntityQueryParser.parse_string('*[layer=="0"]', parse_all=True)
        assert ("layer", "==", "0") == tuple(result.AttribQuery)

    def test_relation_ne(self):
        result = EntityQueryParser.parse_string('*[layer!="0"]', parse_all=True)
        assert ("layer", "!=", "0") == tuple(result.AttribQuery)

    def test_relation_ge(self):
        result = EntityQueryParser.parse_string('*[layer>="0"]', parse_all=True)
        assert ("layer", ">=", "0") == tuple(result.AttribQuery)

    def test_relation_gt(self):
        result = EntityQueryParser.parse_string('*[layer>="0"]', parse_all=True)
        assert ("layer", ">=", "0") == tuple(result.AttribQuery)

    def test_regex_match(self):
        result = EntityQueryParser.parse_string('*[layer?"0"]', parse_all=True)
        assert ("layer", "?", "0") == tuple(result.AttribQuery)

    def test_not_regex_match(self):
        result = EntityQueryParser.parse_string('*[layer!?"0"]', parse_all=True)
        assert ("layer", "!?", "0") == tuple(result.AttribQuery)

    def test_appended_ignore_case_option(self):
        result = EntityQueryParser.parse_string(
            '*[layer=="IgnoreCase"]i', parse_all=True
        )
        assert "i" == result.AttribQueryOptions


class TestInfixBoolQuery:
    def test_not_operation(self):
        result = InfixBoolQuery.parse_string("!a!=1", parse_all=True)
        op, relation = result.AttribQuery
        assert "!" == op
        assert ("a", "!=", 1) == tuple(relation)

    def test_and_operation(self):
        result = InfixBoolQuery.parse_string("a != 1 & b != 2", parse_all=True)
        rel1, op, rel2 = result.AttribQuery
        assert ("a", "!=", 1) == tuple(rel1)
        assert "&" == op
        assert ("b", "!=", 2) == tuple(rel2)

    def test_or_operation(self):
        result = InfixBoolQuery.parse_string("a != 1 | b != 2", parse_all=True)
        rel1, op, rel2 = result.AttribQuery
        assert ("a", "!=", 1) == tuple(rel1)
        assert "|" == op
        assert ("b", "!=", 2) == tuple(rel2)

    def test_not_operation_with_brackets(self):
        result = InfixBoolQuery.parse_string("!(a!=1)", parse_all=True)
        op, relation = result.AttribQuery
        assert "!" == op
        assert ("a", "!=", 1) == tuple(relation)

    def test_operation_with_brackets(self):
        result = InfixBoolQuery.parse_string(
            "(a != 1) & (b != 2)", parse_all=True
        )
        rel1, op, rel2 = result.AttribQuery
        assert ("a", "!=", 1) == tuple(rel1)
        assert "&" == op
        assert ("b", "!=", 2) == tuple(rel2)

    def test_operation_with_nested_brackets(self):
        result = InfixBoolQuery.parse_string(
            "((a != 1) & (b != 2))", parse_all=True
        )
        rel1, op, rel2 = result.AttribQuery
        assert ("a", "!=", 1) == tuple(rel1)
        assert "&" == op
        assert ("b", "!=", 2) == tuple(rel2)



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# created 2019-02-15
import pytest
import ezdxf
from ezdxf.entities.helix import Helix
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

HELIX = """0
HELIX
5
0
330
0
100
AcDbEntity
8
0
100
AcDbSpline
70
0
71
3
72
0
73
0
74
0
100
AcDbHelix
90
29
91
63
10
0.0
20
0.0
30
0.0
11
1.0
21
0.0
31
0.0
12
0.0
22
0.0
32
1.0
40
1.0
41
1.0
42
1.0
290
1
280
1
"""


@pytest.fixture
def entity():
    return Helix.from_text(HELIX)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "HELIX" in ENTITY_CLASSES


def test_default_init():
    entity = Helix()
    assert entity.dxftype() == "HELIX"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = Helix.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": 7,
            "axis_base_point": (1, 2, 3),
            "start_point": (4, 5, 6),
            "axis_vector": (7, 7, 7),
            "radius": 20,
            "turns": 5,
            "handedness": 0,
            "constrain": 2,
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.major_release_number == 29
    assert entity.dxf.maintenance_release_number == 63
    assert entity.dxf.axis_base_point == (1, 2, 3)
    assert entity.dxf.start_point == (4, 5, 6)
    assert entity.dxf.axis_vector == (7, 7, 7)
    assert entity.dxf.radius == 20
    assert entity.dxf.turns == 5
    assert entity.dxf.handedness == 0
    assert entity.dxf.constrain == 2


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.major_release_number == 29
    assert entity.dxf.maintenance_release_number == 63
    assert entity.dxf.axis_base_point == (0, 0, 0)
    assert entity.dxf.start_point == (1, 0, 0)
    assert entity.dxf.axis_vector == (0, 0, 1)
    assert entity.dxf.radius == 1
    assert entity.dxf.turns == 1
    assert entity.dxf.handedness == 1
    assert entity.dxf.constrain == 1


def test_write_dxf():
    entity = Helix.from_text(HELIX)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(HELIX)
    assert result == expected


def test_generic_helix():
    doc = ezdxf.new()
    msp = doc.modelspace()
    helix = msp.new_entity("HELIX", {})
    assert helix.dxftype() == "HELIX"
    assert helix.dxf.major_release_number == 29
    assert helix.dxf.degree == 3



# Copyright (c) 2023, Manfred Moitzi
# License: MIT License

import pytest
import ezdxf
import ezdxf.path
from ezdxf.npshapes import NumpyPath2d
from ezdxf.math import Vec2
from ezdxf.addons.drawing import RenderContext, Frontend
from ezdxf.addons.drawing.recorder import Recorder, BackendProperties, Override, FilledPathsRecord
from ezdxf.addons.drawing.debug_backend import PathBackend


class MyTestFrontend(Frontend):
    def __init__(self, ctx, backend):
        super().__init__(ctx, backend)
        self.out = backend


@pytest.fixture
def doc():
    d = ezdxf.new()
    d.layers.new("Test1")
    d.styles.add("DEJAVU", font="DejaVuSans.ttf")
    return d


@pytest.fixture
def msp(doc):
    return doc.modelspace()


@pytest.fixture
def ctx(doc):
    return RenderContext(doc)


@pytest.fixture
def frontend(doc, ctx):
    return MyTestFrontend(ctx, Recorder())


def replay(frontend, backend):
    player = frontend.out.player().copy()
    player.replay(backend)
    return backend.collector


def test_replay_layout(msp, frontend):
    msp.add_point((0, 0))
    msp.add_point((0, 0))
    frontend.draw_layout(msp)
    bg, pt0, pt1 = replay(frontend, PathBackend())
    assert bg[0] == "bgcolor"
    assert pt0[0] == "point"
    assert pt1[0] == "point"


def test_points(msp, frontend):
    msp.add_point((0, 0), dxfattribs={"layer": "Test1"})
    msp.add_point((0, 0), dxfattribs={"layer": "fail"})
    frontend.draw_entities(msp)
    bg, pt0, pt1 = replay(frontend, PathBackend())
    assert bg[0] == "bgcolor"
    assert pt0[0] == "point"
    assert pt0[-1].layer == "Test1"
    assert pt1[0] == "point"
    assert pt1[-1].layer == "fail"


def test_line(msp, frontend):
    msp.add_line((0, 0), (1, 0))
    frontend.draw_entities(msp)
    _, line = replay(frontend, PathBackend())
    assert line[0] == "line"


def test_lwpolyline_as_path(msp, frontend):
    msp.add_lwpolyline([(0, 0), (1, 0), (2, 0)])
    frontend.draw_entities(msp)
    _, path = replay(frontend, PathBackend())
    assert path[0] == "path"


def test_replay_properties(msp, frontend):
    pline = msp.add_lwpolyline(
        [(0, 0), (1, 0), (2, 0)], dxfattribs={"color": 1, "lineweight": 70}
    )
    frontend.draw_entities(msp)
    _, path = replay(frontend, PathBackend())
    properties = path[2]
    assert properties.color == "#ff0000"
    assert properties.lineweight == pytest.approx(0.7)
    assert properties.layer == "0"
    assert properties.pen == 1
    assert properties.handle == pline.dxf.handle


def test_override_properties_at_replay(msp, ctx):
    def override(_: BackendProperties) -> Override:
        return Override(properties=BackendProperties("#00ff00", 0.5, "1", 2, "FEFE"))

    msp.add_lwpolyline(
        [(0, 0), (1, 0), (2, 0)], dxfattribs={"color": 1, "lineweight": 70}
    )
    # recording:
    backend_recorder = Recorder()
    MyTestFrontend(ctx, backend_recorder).draw_entities(msp)

    # replay:
    player = backend_recorder.player().copy()
    replay_backend = PathBackend()
    player.replay(replay_backend, override)

    properties = replay_backend.collector[1][2]
    assert properties.color == "#00ff00"
    assert properties.lineweight == pytest.approx(0.5)
    assert properties.layer == "1"
    assert properties.pen == 2
    assert properties.handle == "FEFE"


def test_banded_lwpolyline_as_filled_polygon(msp, frontend):
    msp.add_lwpolyline([(0, 0), (1, 0), (2, 0)], dxfattribs={"const_width": 0.1})
    frontend.draw_entities(msp)
    _, filled_polygon = replay(frontend, PathBackend())
    assert filled_polygon[0] == "filled_polygon"


def test_2d_text_as_filled_paths(msp, frontend):
    # since v1.0.4 the frontend does the text rendering and passes only filled
    # polygons to the backend
    msp.add_text(
        "test\ntest", dxfattribs={"style": "DEJAVU"}
    )  # \n shouldn't be  problem. Will be ignored
    frontend.draw_entities(msp)
    _, *filled_paths = replay(frontend, PathBackend())
    assert filled_paths[0][0] == "filled_path"
    text_paths = filled_paths[0][1]
    assert len(text_paths) == 8, "expected a path for each character"


def test_bounding_box(msp, frontend):
    recorder = frontend.out
    assert isinstance(recorder, Recorder)

    msp.add_lwpolyline([(0, 0), (200, 0), (200, 100), (0, 100)])
    frontend.draw_layout(msp)
    player = recorder.player().copy()
    player.replay(PathBackend())
    bbox = player.bbox()
    assert bbox.extmin.isclose((0, 0))
    assert bbox.extmax.isclose((200, 100))


class TestCroppingRecords:
    """Clipping is tested in 822 and 618!"""

    def test_null_sized_crop_box_removes_everything(self):
        """A cropping box of size zero in any dimension should remove all records."""
        props = BackendProperties()
        recorder = Recorder()
        recorder.draw_point(Vec2(0, 0), props)

        # 0 x 0
        player = recorder.player().copy()
        assert len(player.records) == 1
        player.crop_rect((0, 0), (0, 0), 1)
        assert len(player.records) == 0

        # 0 x 1
        player = recorder.player().copy()
        player.crop_rect((0, 0), (0, 1), 1)
        assert len(player.records) == 0

        # 1 x 0
        player = recorder.player().copy()
        player.crop_rect((0, 0), (1, 0), 1)
        assert len(player.records) == 0

    def test_remove_entities_outside(self):
        """Records complete outside the crop box should be removed."""
        props = BackendProperties()
        recorder = Recorder()

        # point outside:
        recorder.draw_point(Vec2(1000, 1000), props)

        # line coincident with crop box side is outside!
        recorder.draw_line(Vec2(100, 0), Vec2(100, 100), props)

        # path coincident with crop box side is outside!
        path = NumpyPath2d.from_vertices(Vec2.list([(0, 100), (50, 100), (100, 100)]))
        recorder.draw_path(path, props)

        player = recorder.player()
        assert len(player.records) == 3

        # crop recordings:
        player.crop_rect((0, 0), (100, 100), 1)

        assert len(player.records) == 0

    def test_entities_inside_crop_box_do_not_change(self):
        """Records complete inside the crop box should not be processed in any kind."""
        props = BackendProperties()
        recorder = Recorder()
        recorder.draw_point(Vec2(50, 50), props)

        player = recorder.player()
        rec0 = player.records[0]
        player.crop_rect((0, 0), (100, 100), 1)

        assert player.records[0] is rec0, "should be identical to original record"

    def test_crop_filled_paths(self):
        props = BackendProperties()
        recorder = Recorder()
        square = NumpyPath2d(ezdxf.path.rect(100, 100))  # center = (0, 0)
        hole = NumpyPath2d(ezdxf.path.rect(50, 50))  # center = (0, 0)
        recorder.draw_filled_paths([square, hole], props)

        player = recorder.player()
        record0 = player.records[0]
        assert isinstance(record0, FilledPathsRecord) 
        paths0 = record0.paths

        player.crop_rect((0, 0), (100, 100), 1)
        record1 = player.records[0]
        assert isinstance(record1, FilledPathsRecord) 
        assert record1.paths is not paths0, "should be new record data"

        paths = record1.paths
        assert paths[0].extents() == ((0, 0), (50, 50)), "should be cropped"
        assert paths[1].extents() == ((0, 0), (25, 25)), "should be cropped"

    def test_does_not_crop_holes_inside_crop_box(self):
        props = BackendProperties()
        recorder = Recorder()
        square = NumpyPath2d(ezdxf.path.rect(100, 100))  # center = (0, 0)
        hole = NumpyPath2d.from_vertices(
            Vec2.list([(10, 10), (20, 10), (20, 20), (10, 20)]), close=True
        )
        recorder.draw_filled_paths([square, hole], props)

        player = recorder.player()
        record0 = player.records[0]
        assert isinstance(record0, FilledPathsRecord) 

        paths0 = record0.paths
        hole0 = paths0[1]
        player.crop_rect((0, 0), (100, 100), 1)

        record1 = player.records[0]
        assert isinstance(record1, FilledPathsRecord) 

        paths = record1.paths
        assert paths[1] is hole0, "should be identical to original hole"

    def test_does_remove_holes_outside_crop_box(self):
        props = BackendProperties()
        recorder = Recorder()

        square = NumpyPath2d(ezdxf.path.rect(100, 100))  # center = (0, 0)
        hole = NumpyPath2d.from_vertices(
            Vec2.list([(-10, -10), (-20, -10), (-20, -20), (-10, -20)]), close=True
        )
        recorder.draw_filled_paths([square, hole], props)

        player = recorder.player()
        player.crop_rect((0, 0), (100, 100), 1)

        record0 = player.records[0]
        assert isinstance(record0, FilledPathsRecord) 

        assert len(record0.paths) == 1, "hole should be removed"



# Copyright (c) 2021-2024 Manfred Moitzi
# License: MIT License
# pylint: disable=unused-variable
from __future__ import annotations
from typing import (
    Iterator,
    Sequence,
    Optional,
    Generic,
    TypeVar,
)
import math

# The pure Python implementation can't import from ._ctypes or ezdxf.math!
from ._vector import Vec3, Vec2
from ._matrix44 import Matrix44


__all__ = ["Bezier3P"]


def check_if_in_valid_range(t: float) -> None:
    if not 0.0 <= t <= 1.0:
        raise ValueError("t not in range [0 to 1]")


T = TypeVar("T", Vec2, Vec3)


class Bezier3P(Generic[T]):
    """Implements an optimized quadratic `Bézier curve`_ for exact 3 control
    points.

    The class supports points of type :class:`Vec2` and :class:`Vec3` as input, the
    class instances are immutable.

    Args:
        defpoints: sequence of definition points as :class:`Vec2` or
            :class:`Vec3` compatible objects.

    """

    __slots__ = ("_control_points", "_offset")

    def __init__(self, defpoints: Sequence[T]):
        if len(defpoints) != 3:
            raise ValueError("Three control points required.")
        point_type = defpoints[0].__class__
        if not point_type.__name__ in ("Vec2", "Vec3"):  # Cython types!!!
            raise TypeError(f"invalid point type: {point_type.__name__}")

        # The start point is the curve offset
        offset: T = defpoints[0]
        self._offset: T = offset
        # moving the curve to the origin reduces floating point errors:
        self._control_points: tuple[T, ...] = tuple(p - offset for p in defpoints)

    @property
    def control_points(self) -> Sequence[T]:
        """Control points as tuple of :class:`Vec3` or :class:`Vec2` objects."""
        # ezdxf optimization: p0 is always (0, 0, 0)
        _, p1, p2 = self._control_points
        offset = self._offset
        return offset, p1 + offset, p2 + offset

    def tangent(self, t: float) -> T:
        """Returns direction vector of tangent for location `t` at the
        Bèzier-curve.

        Args:
            t: curve position in the range ``[0, 1]``

        """
        check_if_in_valid_range(t)
        return self._get_curve_tangent(t)

    def point(self, t: float) -> T:
        """Returns point for location `t` at the Bèzier-curve.

        Args:
            t: curve position in the range ``[0, 1]``

        """
        check_if_in_valid_range(t)
        return self._get_curve_point(t)

    def approximate(self, segments: int) -> Iterator[T]:
        """Approximate `Bézier curve`_ by vertices, yields `segments` + 1
        vertices as ``(x, y[, z])`` tuples.

        Args:
            segments: count of segments for approximation

        """
        if segments < 1:
            raise ValueError(segments)
        delta_t: float = 1.0 / segments
        cp = self.control_points
        yield cp[0]
        for segment in range(1, segments):
            yield self._get_curve_point(delta_t * segment)
        yield cp[2]

    def approximated_length(self, segments: int = 128) -> float:
        """Returns estimated length of Bèzier-curve as approximation by line
        `segments`.
        """
        length: float = 0.0
        prev_point: Optional[T] = None
        for point in self.approximate(segments):
            if prev_point is not None:
                length += prev_point.distance(point)
            prev_point = point
        return length

    def flattening(self, distance: float, segments: int = 4) -> Iterator[T]:
        """Adaptive recursive flattening. The argument `segments` is the
        minimum count of approximation segments, if the distance from the center
        of the approximation segment to the curve is bigger than `distance` the
        segment will be subdivided.

        Args:
            distance: maximum distance from the center of the quadratic (C2)
                curve to the center of the linear (C1) curve between two
                approximation points to determine if a segment should be
                subdivided.
            segments: minimum segment count

        """
        stack: list[tuple[float, T]] = []
        dt: float = 1.0 / segments
        t0: float = 0.0
        t1: float
        cp = self.control_points
        start_point: T = cp[0]
        end_point: T

        yield start_point
        while t0 < 1.0:
            t1 = t0 + dt
            if math.isclose(t1, 1.0):
                end_point = cp[2]
                t1 = 1.0
            else:
                end_point = self._get_curve_point(t1)

            while True:
                mid_t: float = (t0 + t1) * 0.5
                mid_point: T = self._get_curve_point(mid_t)
                chk_point: T = start_point.lerp(end_point)

                d = chk_point.distance(mid_point)
                if d < distance:
                    yield end_point
                    t0 = t1
                    start_point = end_point
                    if stack:
                        t1, end_point = stack.pop()
                    else:
                        break
                else:
                    stack.append((t1, end_point))
                    t1 = mid_t
                    end_point = mid_point

    def _get_curve_point(self, t: float) -> T:
        # 1st control point (p0) is always (0, 0, 0)
        # => p0 * a is always (0, 0, 0)
        _, p1, p2 = self._control_points
        _1_minus_t = 1.0 - t
        # a = (1 - t) ** 2
        b = 2.0 * t * _1_minus_t
        c = t * t
        # add offset at last - it is maybe very large
        return p1 * b + p2 * c + self._offset

    def _get_curve_tangent(self, t: float) -> T:
        # tangent vector is independent from offset location!
        # 1st control point (p0) is always (0, 0, 0)
        # => p0 * a is always (0, 0, 0)
        _, p1, p2 = self._control_points
        # a = -2 * (1 - t)
        b = 2.0 - 4.0 * t
        c = 2.0 * t
        return p1 * b + p2 * c

    def reverse(self) -> Bezier3P[T]:
        """Returns a new Bèzier-curve with reversed control point order."""
        return Bezier3P(list(reversed(self.control_points)))

    def transform(self, m: Matrix44) -> Bezier3P[Vec3]:
        """General transformation interface, returns a new :class:`Bezier3P`
        curve and it is always a 3D curve.

        Args:
             m: 4x4 transformation :class:`Matrix44`

        """
        defpoints = Vec3.generate(self.control_points)
        return Bezier3P(tuple(m.transform_vertices(defpoints)))



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Iterable, TYPE_CHECKING, no_type_check
from functools import lru_cache
import enum
import numpy as np

from ezdxf import colors
from ezdxf.lldxf.const import VALID_DXF_LINEWEIGHTS
from ezdxf.math import Vec2, BoundingBox2d, Matrix44
from ezdxf.path import to_splines_and_polylines, to_hatches
from ezdxf.layouts import BaseLayout

from .type_hints import Color
from .backend import BackendInterface, BkPath2d, BkPoints2d, ImageData
from .config import Configuration
from .properties import BackendProperties

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.entities import Solid


class ColorMode(enum.Enum):
    """This enum is used to define the color output mode of the :class:`DXFBackend`.

    Attributes:
        ACI: the color is set as :ref:`ACI` and assigned by layer
        RGB: the color is set as RGB true color value

    """

    # Use color index as primary color
    ACI = enum.auto()

    # Use always the RGB value
    RGB = enum.auto()


DARK_COLOR_THRESHOLD = 0.2
RGB_BLACK = colors.RGB(0, 0, 0)
BYLAYER = 256


class DXFBackend(BackendInterface):
    """The :class:`DXFBackend` creates simple DXF files of POINT, LINE, LWPOLYLINE and
    HATCH entities. This backend does ot need any additional packages.

    Args:
        layout: a DXF :class:`~ezdxf.layouts.BaseLayout`
        color_mode: see :class:`ColorMode`

    """

    def __init__(
        self, layout: BaseLayout, color_mode: ColorMode = ColorMode.RGB
    ) -> None:
        assert layout.doc is not None, "valid DXF document required"
        super().__init__()
        self.layout = layout
        self.doc = layout.doc
        self.color_mode = color_mode
        self.bg_color = RGB_BLACK
        self.is_dark_bg = True
        self._layers: dict[int, str] = dict()
        self._dxfattribs: dict[int, dict] = dict()

    def set_background(self, color: Color) -> None:
        self.bg_color = colors.RGB.from_hex(color)
        self.is_dark_bg = self.bg_color.luminance < DARK_COLOR_THRESHOLD

    def get_layer_name(self, pen: int) -> str:
        try:
            return self._layers[pen]
        except KeyError:
            pass

        layer_name = f"PEN_{pen:03d}"
        self._layers[pen] = layer_name
        if not self.doc.layers.has_entry(layer_name):
            self.doc.layers.add(layer_name, color=pen)
        return layer_name

    def resolve_properties(self, properties: BackendProperties) -> dict:
        key = hash(properties)
        try:
            return self._dxfattribs[key]
        except KeyError:
            pass

        rgb = properties.rgb
        pen = properties.pen
        if pen < 1 or pen > 255:
            pen = 7
        aci = pen
        if self.color_mode == ColorMode.ACI:
            aci = BYLAYER
        attribs = {
            "color": aci,
            "layer": self.get_layer_name(pen),
            "lineweight": make_lineweight(properties.lineweight),
        }
        if self.color_mode == ColorMode.RGB:
            attribs["true_color"] = colors.rgb2int(rgb)

        alpha = properties.color[7:9]
        if alpha:
            try:
                f = int(alpha, 16) / 255
            except ValueError:
                pass
            else:
                attribs["transparency"] = colors.float2transparency(f)
        self._dxfattribs[key] = attribs
        return attribs

    def set_solid_fill(self, hatch, properties: BackendProperties) -> None:
        rgb: colors.RGB | None = None
        aci = BYLAYER
        if self.color_mode == ColorMode.RGB:
            rgb = properties.rgb
            aci = properties.pen
        hatch.set_solid_fill(color=aci, style=0, rgb=rgb)

    def draw_point(self, pos: Vec2, properties: BackendProperties) -> None:
        self.layout.add_point(pos, dxfattribs=self.resolve_properties(properties))

    def draw_line(self, start: Vec2, end: Vec2, properties: BackendProperties) -> None:
        self.layout.add_line(start, end, dxfattribs=self.resolve_properties(properties))

    def draw_solid_lines(
        self, lines: Iterable[tuple[Vec2, Vec2]], properties: BackendProperties
    ) -> None:
        lines = list(lines)
        if len(lines) == 0:
            return
        attribs = self.resolve_properties(properties)
        for start, end in lines:
            self.layout.add_line(start, end, dxfattribs=attribs)

    def draw_path(self, path: BkPath2d, properties: BackendProperties) -> None:
        attribs = self.resolve_properties(properties)
        if path.has_curves:
            for entity in to_splines_and_polylines(path, dxfattribs=attribs):  # type: ignore
                self.layout.add_entity(entity)
        else:
            self.layout.add_lwpolyline(path.control_vertices(), dxfattribs=attribs)

    def draw_filled_paths(
        self, paths: Iterable[BkPath2d], properties: BackendProperties
    ) -> None:
        attribs = self.resolve_properties(properties)
        py_paths = [p.to_path() for p in paths]
        for hatch in to_hatches(py_paths, dxfattribs=attribs):
            self.layout.add_entity(hatch)
            self.set_solid_fill(hatch, properties)

    def draw_filled_polygon(
        self, points: BkPoints2d, properties: BackendProperties
    ) -> None:
        hatch = self.layout.add_hatch(dxfattribs=self.resolve_properties(properties))
        hatch.paths.add_polyline_path(points.vertices(), is_closed=True)
        self.set_solid_fill(hatch, properties)

    def draw_image(self, image_data: ImageData, properties: BackendProperties) -> None:
        pass  # TODO: not implemented

    def configure(self, config: Configuration) -> None:
        pass

    def clear(self) -> None:
        pass

    def finalize(self) -> None:
        pass

    def enter_entity(self, entity, properties) -> None:
        pass

    def exit_entity(self, entity) -> None:
        pass


def alpha_to_transparency(alpha: int) -> float:
    return colors.float2transparency(alpha / 255)


@lru_cache(maxsize=None)
def make_lineweight(width: float) -> int:
    width_int = int(width * 100)
    for lw in VALID_DXF_LINEWEIGHTS:
        if width_int <= lw:
            return lw
    return VALID_DXF_LINEWEIGHTS[-1]


@no_type_check
def update_extents(doc: Drawing, bbox: BoundingBox2d) -> None:
    doc.header["$EXTMIN"] = (bbox.extmin.x, bbox.extmin.y, 0)
    doc.header["$EXTMAX"] = (bbox.extmax.x, bbox.extmax.y, 0)


def setup_paperspace(doc: Drawing, bbox: BoundingBox2d):
    psp_size = bbox.size / 40.0  # plu to mm
    psp_center = psp_size * 0.5
    psp = doc.paperspace()
    psp.page_setup(size=(psp_size.x, psp_size.y), margins=(0, 0, 0, 0), units="mm")
    psp.add_viewport(
        center=psp_center,
        size=(psp_size.x, psp_size.y),
        view_center_point=bbox.center,
        view_height=bbox.size.y,
        status=2,
    )


def add_background(msp: BaseLayout, bbox: BoundingBox2d, color: colors.RGB) -> Solid:
    v = bbox.rect_vertices()
    bg = msp.add_solid(
        [v[0], v[1], v[3], v[2]], dxfattribs={"true_color": colors.rgb2int(color)}
    )
    return bg



#!/usr/bin/env python3
# Copyright (c) 2020-2023, Matthew Broadway
# License: MIT License
# mypy: ignore_errors=True
from __future__ import annotations
from typing import Iterable, Sequence, Set, Optional
import math
import os
import time

from ezdxf.addons.xqt import QtWidgets as qw, QtCore as qc, QtGui as qg
from ezdxf.addons.xqt import Slot, QAction, Signal

import ezdxf
import ezdxf.bbox
from ezdxf import recover
from ezdxf.addons import odafc
from ezdxf.addons.drawing import Frontend, RenderContext
from ezdxf.addons.drawing.config import Configuration

from ezdxf.addons.drawing.properties import (
    is_dark_color,
    set_layers_state,
    LayerProperties,
)
from ezdxf.addons.drawing.pyqt import (
    _get_x_scale,
    PyQtBackend,
    CorrespondingDXFEntity,
    CorrespondingDXFParentStack,
)
from ezdxf.audit import Auditor
from ezdxf.document import Drawing
from ezdxf.entities import DXFGraphic, DXFEntity
from ezdxf.layouts import Layout
from ezdxf.lldxf.const import DXFStructureError


class CADGraphicsView(qw.QGraphicsView):
    closing = Signal()

    def __init__(
        self,
        *,
        view_buffer: float = 0.2,
        zoom_per_scroll_notch: float = 0.2,
        loading_overlay: bool = True,
    ):
        super().__init__()
        self._zoom = 1.0
        self._default_zoom = 1.0
        self._zoom_limits = (0.5, 100)
        self._zoom_per_scroll_notch = zoom_per_scroll_notch
        self._view_buffer = view_buffer
        self._loading_overlay = loading_overlay
        self._is_loading = False

        self.setTransformationAnchor(qw.QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(qw.QGraphicsView.AnchorUnderMouse)
        self.setVerticalScrollBarPolicy(qc.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(qc.Qt.ScrollBarAlwaysOff)
        self.setDragMode(qw.QGraphicsView.ScrollHandDrag)
        self.setFrameShape(qw.QFrame.NoFrame)
        self.setRenderHints(
            qg.QPainter.Antialiasing
            | qg.QPainter.TextAntialiasing
            | qg.QPainter.SmoothPixmapTransform
        )

        self.setScene(qw.QGraphicsScene())
        self.scale(1, -1)  # so that +y is up

    def closeEvent(self, event: qg.QCloseEvent) -> None:
        super().closeEvent(event)
        self.closing.emit()

    def clear(self):
        pass

    def begin_loading(self):
        self._is_loading = True
        self.scene().invalidate(qc.QRectF(), qw.QGraphicsScene.AllLayers)
        qw.QApplication.processEvents()

    def end_loading(self, new_scene: qw.QGraphicsScene):
        self.setScene(new_scene)
        self._is_loading = False
        self.buffer_scene_rect()
        self.scene().invalidate(qc.QRectF(), qw.QGraphicsScene.AllLayers)

    def buffer_scene_rect(self):
        scene = self.scene()
        r = scene.sceneRect()
        bx, by = (
            r.width() * self._view_buffer / 2,
            r.height() * self._view_buffer / 2,
        )
        scene.setSceneRect(r.adjusted(-bx, -by, bx, by))

    def fit_to_scene(self):
        self.fitInView(self.sceneRect(), qc.Qt.KeepAspectRatio)
        self._default_zoom = _get_x_scale(self.transform())
        self._zoom = 1

    def _get_zoom_amount(self) -> float:
        return _get_x_scale(self.transform()) / self._default_zoom

    def wheelEvent(self, event: qg.QWheelEvent) -> None:
        # dividing by 120 gets number of notches on a typical scroll wheel.
        # See QWheelEvent documentation
        delta_notches = event.angleDelta().y() / 120
        direction = math.copysign(1, delta_notches)
        factor = (1.0 + self._zoom_per_scroll_notch * direction) ** abs(delta_notches)
        resulting_zoom = self._zoom * factor
        if resulting_zoom < self._zoom_limits[0]:
            factor = self._zoom_limits[0] / self._zoom
        elif resulting_zoom > self._zoom_limits[1]:
            factor = self._zoom_limits[1] / self._zoom
        self.scale(factor, factor)
        self._zoom *= factor

    def save_view(self) -> SavedView:
        return SavedView(
            self.transform(),
            self._default_zoom,
            self._zoom,
            self.horizontalScrollBar().value(),
            self.verticalScrollBar().value(),
        )

    def restore_view(self, view: SavedView):
        self.setTransform(view.transform)
        self._default_zoom = view.default_zoom
        self._zoom = view.zoom
        self.horizontalScrollBar().setValue(view.x)
        self.verticalScrollBar().setValue(view.y)

    def drawForeground(self, painter: qg.QPainter, rect: qc.QRectF) -> None:
        if self._is_loading and self._loading_overlay:
            painter.save()
            painter.fillRect(rect, qg.QColor("#aa000000"))
            painter.setWorldMatrixEnabled(False)
            r = self.viewport().rect()
            painter.setBrush(qc.Qt.NoBrush)
            painter.setPen(qc.Qt.white)
            painter.drawText(r.center(), "Loading...")
            painter.restore()


class SavedView:
    def __init__(
        self, transform: qg.QTransform, default_zoom: float, zoom: float, x: int, y: int
    ):
        self.transform = transform
        self.default_zoom = default_zoom
        self.zoom = zoom
        self.x = x
        self.y = y


class CADGraphicsViewWithOverlay(CADGraphicsView):
    mouse_moved = Signal(qc.QPointF)
    element_hovered = Signal(object, int)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._selected_items: list[qw.QGraphicsItem] = []
        self._selected_index = None
        self._mark_selection = True

    @property
    def current_hovered_element(self) -> Optional[DXFEntity]:
        if self._selected_items:
            graphics_item = self._selected_items[self._selected_index]
            dxf_entity = graphics_item.data(CorrespondingDXFEntity)
            return dxf_entity
        else:
            return None

    def clear(self):
        super().clear()
        self._selected_items = None
        self._selected_index = None

    def begin_loading(self):
        self.clear()
        super().begin_loading()

    def drawForeground(self, painter: qg.QPainter, rect: qc.QRectF) -> None:
        super().drawForeground(painter, rect)
        if self._selected_items and self._mark_selection:
            item = self._selected_items[self._selected_index]
            r = item.sceneTransform().mapRect(item.boundingRect())
            painter.fillRect(r, qg.QColor(0, 255, 0, 100))

    def mouseMoveEvent(self, event: qg.QMouseEvent) -> None:
        super().mouseMoveEvent(event)
        pos = self.mapToScene(event.pos())
        self.mouse_moved.emit(pos)
        selected_items = self.scene().items(pos)
        if selected_items != self._selected_items:
            self._selected_items = selected_items
            self._selected_index = 0 if self._selected_items else None
            self._emit_selected()

    def mouseReleaseEvent(self, event: qg.QMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        if event.button() == qc.Qt.LeftButton and self._selected_items:
            self._selected_index = (self._selected_index + 1) % len(
                self._selected_items
            )
            self._emit_selected()

    def _emit_selected(self):
        self.element_hovered.emit(self._selected_items, self._selected_index)
        self.scene().invalidate(self.sceneRect(), qw.QGraphicsScene.ForegroundLayer)

    def toggle_selection_marker(self):
        self._mark_selection = not self._mark_selection


class CADWidget(qw.QWidget):
    def __init__(self, view: CADGraphicsView, config: Configuration = Configuration()):
        super().__init__()
        layout = qw.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(view)
        self.setLayout(layout)
        self._view = view
        self._view.closing.connect(self.close)
        self._config = config
        self._bbox_cache = ezdxf.bbox.Cache()
        self._doc: Drawing = None  # type: ignore
        self._render_context: RenderContext = None  # type: ignore
        self._visible_layers: set[str] = set()
        self._current_layout: str = "Model"
        self._reset_backend()

    def _reset_backend(self):
        # clear caches
        self._backend = PyQtBackend()

    @property
    def doc(self) -> Drawing:
        return self._doc

    @property
    def view(self) -> CADGraphicsView:
        return self._view

    @property
    def render_context(self) -> RenderContext:
        return self._render_context

    @property
    def current_layout(self) -> str:
        return self._current_layout

    def set_document(
        self,
        document: Drawing,
        *,
        layout: str = "Model",
        draw: bool = True,
    ):
        self._doc = document
        # initialize bounding box cache for faste paperspace drawing
        self._bbox_cache = ezdxf.bbox.Cache()
        self._render_context = self._make_render_context(document)
        self._reset_backend()
        self._visible_layers = set()
        self._current_layout = None
        if draw:
            self.draw_layout(layout)

    def set_visible_layers(self, layers: Set[str]) -> None:
        self._visible_layers = layers
        self.draw_layout(self._current_layout, reset_view=False)

    def _make_render_context(self, doc: Drawing) -> RenderContext:
        def update_layers_state(layers: Sequence[LayerProperties]):
            if self._visible_layers:
                set_layers_state(layers, self._visible_layers, state=True)

        render_context = RenderContext(doc)
        render_context.set_layer_properties_override(update_layers_state)
        return render_context

    def draw_layout(
        self,
        layout_name: str,
        reset_view: bool = True,
    ):
        self._current_layout = layout_name
        self._view.begin_loading()
        new_scene = qw.QGraphicsScene()
        self._backend.set_scene(new_scene)
        layout = self._doc.layout(layout_name)
        self._update_render_context(layout)
        try:
            self._create_frontend().draw_layout(layout)
        finally:
            self._backend.finalize()
        self._view.end_loading(new_scene)
        self._view.buffer_scene_rect()
        if reset_view:
            self._view.fit_to_scene()

    def _create_frontend(self) -> Frontend:
        return Frontend(
            ctx=self._render_context,
            out=self._backend,
            config=self._config,
            bbox_cache=self._bbox_cache,
        )

    def _update_render_context(self, layout: Layout) -> None:
        assert self._render_context is not None
        self._render_context.set_current_layout(layout)


class CADViewer(qw.QMainWindow):
    def __init__(self, cad: Optional[CADWidget] = None):
        super().__init__()
        self._doc: Optional[Drawing] = None
        if cad is None:
            self._cad = CADWidget(CADGraphicsViewWithOverlay(), config=Configuration())
        else:
            self._cad = cad
        self._view = self._cad.view

        if isinstance(self._view, CADGraphicsViewWithOverlay):
            self._view.element_hovered.connect(self._on_element_hovered)
            self._view.mouse_moved.connect(self._on_mouse_moved)

        menu = self.menuBar()
        select_doc_action = QAction("Select Document", self)
        select_doc_action.triggered.connect(self._select_doc)
        menu.addAction(select_doc_action)
        self.select_layout_menu = menu.addMenu("Select Layout")

        toggle_sidebar_action = QAction("Toggle Sidebar", self)
        toggle_sidebar_action.triggered.connect(self._toggle_sidebar)
        menu.addAction(toggle_sidebar_action)

        toggle_selection_marker_action = QAction("Toggle Entity Marker", self)
        toggle_selection_marker_action.triggered.connect(self._toggle_selection_marker)
        menu.addAction(toggle_selection_marker_action)

        self.reload_menu = menu.addMenu("Reload")
        reload_action = QAction("Reload", self)
        reload_action.setShortcut(qg.QKeySequence("F5"))
        reload_action.triggered.connect(self._reload)
        self.reload_menu.addAction(reload_action)
        self.keep_view_action = QAction("Keep View", self)
        self.keep_view_action.setCheckable(True)
        self.keep_view_action.setChecked(True)
        self.reload_menu.addAction(self.keep_view_action)
        watch_action = QAction("Watch", self)
        watch_action.setCheckable(True)
        watch_action.toggled.connect(self._toggle_watch)
        self.reload_menu.addAction(watch_action)
        self._watch_timer = qc.QTimer()
        self._watch_timer.setInterval(50)
        self._watch_timer.timeout.connect(self._check_watch)
        self._watch_mtime = None

        self.sidebar = qw.QSplitter(qc.Qt.Vertical)
        self.layers = qw.QListWidget()
        self.layers.setStyleSheet(
            "QListWidget {font-size: 12pt;} "
            "QCheckBox {font-size: 12pt; padding-left: 5px;}"
        )
        self.sidebar.addWidget(self.layers)
        info_container = qw.QWidget()
        info_layout = qw.QVBoxLayout()
        info_layout.setContentsMargins(0, 0, 0, 0)
        self.selected_info = qw.QPlainTextEdit()
        self.selected_info.setReadOnly(True)
        info_layout.addWidget(self.selected_info)
        self.mouse_pos = qw.QLabel()
        info_layout.addWidget(self.mouse_pos)
        info_container.setLayout(info_layout)
        self.sidebar.addWidget(info_container)

        container = qw.QSplitter()
        self.setCentralWidget(container)
        container.addWidget(self._cad)
        container.addWidget(self.sidebar)
        container.setCollapsible(0, False)
        container.setCollapsible(1, True)
        w = container.width()
        container.setSizes([int(3 * w / 4), int(w / 4)])
        self.setWindowTitle("CAD Viewer")
        self.resize(1600, 900)
        self.show()

    @staticmethod
    def from_config(config: Configuration) -> CADViewer:
        return CADViewer(cad=CADWidget(CADGraphicsViewWithOverlay(), config=config))

    def _create_cad_widget(self):
        self._view = CADGraphicsViewWithOverlay()
        self._cad = CADWidget(self._view)

    def load_file(self, path: str, layout: str = "Model"):
        try:
            if os.path.splitext(path)[1].lower() == ".dwg":
                doc = odafc.readfile(path)
                auditor = doc.audit()
            else:
                try:
                    doc = ezdxf.readfile(path)
                except ezdxf.DXFError:
                    doc, auditor = recover.readfile(path)
                else:
                    auditor = doc.audit()
            self.set_document(doc, auditor, layout=layout)
        except IOError as e:
            qw.QMessageBox.critical(self, "Loading Error", str(e))
        except DXFStructureError as e:
            qw.QMessageBox.critical(
                self,
                "DXF Structure Error",
                f'Invalid DXF file "{path}": {str(e)}',
            )

    def _select_doc(self):
        path, _ = qw.QFileDialog.getOpenFileName(
            self,
            caption="Select CAD Document",
            filter="CAD Documents (*.dxf *.DXF *.dwg *.DWG)",
        )
        if path:
            self.load_file(path)

    def set_document(
        self,
        document: Drawing,
        auditor: Auditor,
        *,
        layout: str = "Model",
        draw: bool = True,
    ):
        error_count = len(auditor.errors)
        if error_count > 0:
            ret = qw.QMessageBox.question(
                self,
                "Found DXF Errors",
                f'Found {error_count} errors in file "{document.filename}"\n'
                f"Load file anyway? ",
            )
            if ret == qw.QMessageBox.No:
                auditor.print_error_report(auditor.errors)
                return

        if document.filename:
            try:
                self._watch_mtime = os.stat(document.filename).st_mtime
            except OSError:
                self._watch_mtime = None
        else:
            self._watch_mtime = None
        self._cad.set_document(document, layout=layout, draw=draw)
        self._doc = document
        self._populate_layouts()
        self._populate_layer_list()
        self.setWindowTitle("CAD Viewer - " + str(document.filename))

    def _populate_layer_list(self):
        self.layers.blockSignals(True)
        self.layers.clear()
        for layer in self._cad.render_context.layers.values():
            name = layer.layer
            item = qw.QListWidgetItem()
            self.layers.addItem(item)
            checkbox = qw.QCheckBox(name)
            checkbox.setCheckState(
                qc.Qt.Checked if layer.is_visible else qc.Qt.Unchecked
            )
            checkbox.stateChanged.connect(self._layers_updated)
            text_color = "#FFFFFF" if is_dark_color(layer.color, 0.4) else "#000000"
            checkbox.setStyleSheet(
                f"color: {text_color}; background-color: {layer.color}"
            )
            self.layers.setItemWidget(item, checkbox)
        self.layers.blockSignals(False)

    def _populate_layouts(self):
        def draw_layout(name: str):
            def run():
                self.draw_layout(name, reset_view=True)

            return run

        self.select_layout_menu.clear()
        for layout_name in self._cad.doc.layout_names_in_taborder():
            action = QAction(layout_name, self)
            action.triggered.connect(draw_layout(layout_name))
            self.select_layout_menu.addAction(action)

    def draw_layout(
        self,
        layout_name: str,
        reset_view: bool = True,
    ):
        print(f"drawing {layout_name}")
        try:
            start = time.perf_counter()
            self._cad.draw_layout(layout_name, reset_view=reset_view)
            duration = time.perf_counter() - start
            print(f"took {duration:.4f} seconds")
        except DXFStructureError as e:
            qw.QMessageBox.critical(
                self,
                "DXF Structure Error",
                f'Abort rendering of layout "{layout_name}": {str(e)}',
            )

    def resizeEvent(self, event: qg.QResizeEvent) -> None:
        self._view.fit_to_scene()

    def _layer_checkboxes(self) -> Iterable[tuple[int, qw.QCheckBox]]:
        for i in range(self.layers.count()):
            item = self.layers.itemWidget(self.layers.item(i))
            yield i, item  # type: ignore

    @Slot(int)  # type: ignore
    def _layers_updated(self, item_state: qc.Qt.CheckState):
        shift_held = qw.QApplication.keyboardModifiers() & qc.Qt.ShiftModifier
        if shift_held:
            for i, item in self._layer_checkboxes():
                item.blockSignals(True)
                item.setCheckState(item_state)
                item.blockSignals(False)

        visible_layers = set()
        for i, layer in self._layer_checkboxes():
            if layer.checkState() == qc.Qt.Checked:
                visible_layers.add(layer.text())
        self._cad.set_visible_layers(visible_layers)

    @Slot()
    def _toggle_sidebar(self):
        self.sidebar.setHidden(not self.sidebar.isHidden())

    @Slot()
    def _toggle_selection_marker(self):
        self._view.toggle_selection_marker()

    @Slot()
    def _reload(self):
        if self._cad.doc is not None and self._cad.doc.filename:
            keep_view = self.keep_view_action.isChecked()
            view = self._view.save_view() if keep_view else None
            self.load_file(self._cad.doc.filename, layout=self._cad.current_layout)
            if keep_view:
                self._view.restore_view(view)

    @Slot()
    def _toggle_watch(self):
        if self._watch_timer.isActive():
            self._watch_timer.stop()
        else:
            self._watch_timer.start()

    @Slot()
    def _check_watch(self):
        if self._watch_mtime is None or self._cad.doc is None:
            return
        filename = self._cad.doc.filename
        if filename:
            try:
                mtime = os.stat(filename).st_mtime
            except OSError:
                return
            if mtime != self._watch_mtime:
                self._reload()

    @Slot(qc.QPointF)
    def _on_mouse_moved(self, mouse_pos: qc.QPointF):
        self.mouse_pos.setText(
            f"mouse position: {mouse_pos.x():.4f}, {mouse_pos.y():.4f}\n"
        )

    @Slot(object, int)
    def _on_element_hovered(self, elements: list[qw.QGraphicsItem], index: int):
        if not elements:
            text = "No element selected"
        else:
            text = f"Selected: {index + 1} / {len(elements)}    (click to cycle)\n"
            element = elements[index]
            dxf_entity: DXFGraphic | str | None = element.data(CorrespondingDXFEntity)
            if isinstance(dxf_entity, str):
                dxf_entity = self.load_dxf_entity(dxf_entity)
            if dxf_entity is None:
                text += "No data"
            else:
                text += (
                    f"Selected Entity: {dxf_entity}\n"
                    f"Layer: {dxf_entity.dxf.layer}\n\nDXF Attributes:\n"
                )
                text += _entity_attribs_string(dxf_entity)

                dxf_parent_stack = element.data(CorrespondingDXFParentStack)
                if dxf_parent_stack:
                    text += "\nParents:\n"
                    for entity in reversed(dxf_parent_stack):
                        text += f"- {entity}\n"
                        text += _entity_attribs_string(entity, indent="    ")
        self.selected_info.setPlainText(text)

    def load_dxf_entity(self, entity_handle: str) -> DXFGraphic | None:
        if self._doc is not None:
            return self._doc.entitydb.get(entity_handle)
        return None


def _entity_attribs_string(dxf_entity: DXFGraphic, indent: str = "") -> str:
    text = ""
    for key, value in dxf_entity.dxf.all_existing_dxf_attribs().items():
        text += f"{indent}- {key}: {value}\n"
    return text



# Copyright (c) 2018-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    Iterable,
    Tuple,
    List,
    Sequence,
    Any,
    cast,
    Optional,
)
from typing_extensions import TypeAlias
import abc

from functools import lru_cache
from itertools import repeat
import math
import reprlib

import numpy as np
import numpy.typing as npt
from ezdxf.acc import USE_C_EXT


__all__ = [
    "Matrix",
    "Solver",
    "numpy_vector_solver",
    "numpy_matrix_solver",
    "NumpySolver",
    "tridiagonal_vector_solver",
    "tridiagonal_matrix_solver",
    "detect_banded_matrix",
    "compact_banded_matrix",
    "BandedMatrixLU",
    "banded_matrix",
    "quadratic_equation",
    "cubic_equation",
    "binomial_coefficient",
]


def zip_to_list(*args) -> Iterable[list]:
    for e in zip(*args):  # returns immutable tuples
        yield list(e)  # need mutable list


MatrixData: TypeAlias = List[List[float]]
FrozenMatrixData: TypeAlias = Tuple[Tuple[float, ...]]
Shape: TypeAlias = Tuple[int, int]
NDArray: TypeAlias = npt.NDArray[np.float64]


def copy_float_matrix(A) -> MatrixData:
    if isinstance(A, Matrix):
        A = A.matrix
    return [[float(v) for v in row] for row in A]


@lru_cache(maxsize=128)
def binomial_coefficient(k: int, i: int) -> float:
    # (c) Onur Rauf Bingol <orbingol@gmail.com>, NURBS-Python, MIT-License
    """Computes the binomial coefficient (denoted by `k choose i`).

    Please see the following website for details:
    http://mathworld.wolfram.com/BinomialCoefficient.html

    Args:
        k: size of the set of distinct elements
        i: size of the subsets

    """
    # Special case
    if i > k:
        return float(0)
    # Compute binomial coefficient
    k_fact: int = math.factorial(k)
    i_fact: int = math.factorial(i)
    k_i_fact: int = math.factorial(k - i)
    return float(k_fact / (k_i_fact * i_fact))


class Matrix:
    """Basic matrix implementation based :class:`numpy.ndarray`. Matrix data is stored in
    row major order, this means in a list of rows, where each row is a list of floats.

    Initialization:

        - Matrix(shape=(rows, cols)) ... new matrix filled with zeros
        - Matrix(matrix[, shape=(rows, cols)]) ... from copy of matrix and optional reshape
        - Matrix([[row_0], [row_1], ..., [row_n]]) ... from Iterable[Iterable[float]]
        - Matrix([a1, a2, ..., an], shape=(rows, cols)) ... from Iterable[float] and shape

    .. versionchanged:: 1.2
        Implementation based on :class:`numpy.ndarray`.

    Attributes:
        matrix: matrix data as :class:`numpy.ndarray`

    """

    __slots__ = ("matrix", "abs_tol")

    def __init__(
        self,
        items: Any = None,
        shape: Optional[Shape] = None,
        matrix: Optional[MatrixData | NDArray] = None,
    ):
        self.abs_tol: float = 1e-12
        self.matrix: NDArray = np.array((), dtype=np.float64)
        if matrix is not None:
            self.matrix = np.array(matrix, dtype=np.float64)
            return

        if items is None:
            if shape is not None:
                self.matrix = np.zeros(shape)
            else:  # items is None, shape is None
                return
        elif isinstance(items, Matrix):
            if shape is None:
                shape = items.shape
            self.matrix = items.matrix.reshape(shape).copy()
        else:
            items = list(items)
            try:
                self.matrix = np.array([list(row) for row in items], dtype=np.float64)
            except TypeError:
                if shape is not None:
                    self.matrix = np.array(list(items), dtype=np.float64).reshape(shape)

    def __iter__(self) -> NDArray:
        return np.ravel(self.matrix)

    def __copy__(self) -> "Matrix":
        m = Matrix(matrix=self.matrix.copy())
        m.abs_tol = self.abs_tol
        return m

    def __str__(self) -> str:
        return str(self.matrix)

    def __repr__(self) -> str:
        return f"Matrix({reprlib.repr(self.matrix)})"

    @staticmethod
    def reshape(items: Iterable[float], shape: Shape) -> Matrix:
        """Returns a new matrix for iterable `items` in the configuration of
        `shape`.
        """
        return Matrix(matrix=np.array(list(items), dtype=np.float64).reshape(shape))

    @property
    def nrows(self) -> int:
        """Count of matrix rows."""
        return self.matrix.shape[0]

    @property
    def ncols(self) -> int:
        """Count of matrix columns."""
        return self.matrix.shape[1]

    @property
    def shape(self) -> Shape:
        """Shape of matrix as (n, m) tuple for n rows and m columns."""
        return self.matrix.shape  # type: ignore

    def row(self, index: int) -> list[float]:
        """Returns row `index` as list of floats."""
        return list(self.matrix[index])

    def col(self, index: int) -> list[float]:
        """Return column `index` as list of floats."""
        return list(self.matrix[:, index])

    def diag(self, index: int) -> list[float]:
        """Returns diagonal `index` as list of floats.

        An `index` of 0 specifies the main diagonal, negative values
        specifies diagonals below the main diagonal and positive values
        specifies diagonals above the main diagonal.

        e.g. given a 4x4 matrix:

        - index 0 is [00, 11, 22, 33],
        - index -1 is [10, 21, 32] and
        - index +1 is [01, 12, 23]

        """
        return list(self.matrix.diagonal(index))

    def rows(self) -> list[list[float]]:
        """Return a list of all rows."""
        return list(list(r) for r in self.matrix)

    def cols(self) -> list[list[float]]:
        """Return a list of all columns."""
        return [list(self.col(i)) for i in range(self.ncols)]

    def set_row(self, index: int, items: float | Iterable[float] = 1.0) -> None:
        """Set row values to a fixed value or from an iterable of floats."""
        if isinstance(items, (float, int)):
            items = [float(items)] * self.ncols
        items = list(items)
        if len(items) != self.ncols:
            raise ValueError("Invalid item count")
        self.matrix[index] = items

    def set_col(self, index: int, items: float | Iterable[float] = 1.0) -> None:
        """Set column values to a fixed value or from an iterable of floats."""
        if isinstance(items, (float, int)):
            items = [float(items)] * self.nrows
        self.matrix[:, index] = list(items)

    def set_diag(self, index: int = 0, items: float | Iterable[float] = 1.0) -> None:
        """Set diagonal values to a fixed value or from an iterable of floats.

        An `index` of ``0`` specifies the main diagonal, negative values
        specifies diagonals below the main diagonal and positive values
        specifies diagonals above the main diagonal.

        e.g. given a 4x4 matrix:
        index ``0`` is [00, 11, 22, 33],
        index ``-1`` is [10, 21, 32] and
        index ``+1`` is [01, 12, 23]

        """
        if isinstance(items, (float, int)):
            items = repeat(float(items))

        col_offset: int = max(index, 0)
        row_offset: int = abs(min(index, 0))

        for index, value in zip(range(max(self.nrows, self.ncols)), items):
            try:
                self.matrix[index + row_offset, index + col_offset] = value
            except IndexError:
                return

    @classmethod
    def identity(cls, shape: Shape) -> Matrix:
        """Returns the identity matrix for configuration `shape`."""
        m = Matrix(shape=shape)
        m.set_diag(0, 1.0)
        return m

    def append_row(self, items: Sequence[float]) -> None:
        """Append a row to the matrix."""
        if self.matrix.size == 0:
            self.matrix = np.array([items], dtype=np.float64)
        elif len(items) == self.ncols:
            self.matrix = np.r_[self.matrix, items]
        else:
            raise ValueError("Invalid item count.")

    def append_col(self, items: Sequence[float]) -> None:
        """Append a column to the matrix."""
        if self.matrix.size == 0:
            self.matrix = np.array([[item] for item in items], dtype=np.float64)
        elif len(items) == self.nrows:
            self.matrix = np.c_[self.matrix, items]
        else:
            raise ValueError("Invalid item count.")

    def freeze(self) -> Matrix:
        """Returns a frozen matrix, all data is stored in immutable tuples."""
        m = self.__copy__()
        m.matrix.flags.writeable = False
        return m

    def __getitem__(self, item: tuple[int, int]) -> float:
        """Get value by (row, col) index tuple, fancy slicing as known from
        numpy is not supported.

        """
        return float(self.matrix[item])

    def __setitem__(self, item: tuple[int, int], value: float):
        """Set value by (row, col) index tuple, fancy slicing as known from
        numpy is not supported.

        """
        self.matrix[item] = value

    def __eq__(self, other: object) -> bool:
        """Returns ``True`` if matrices are equal."""
        if not isinstance(other, Matrix):
            raise TypeError("Matrix class required.")
        if self.shape != other.shape:
            raise TypeError("Matrices have different shapes.")
        return bool(np.all(self.matrix == other.matrix))

    def isclose(self, other: object) -> bool:
        """Returns ``True`` if matrices are close to equal, tolerance value for
        comparison is adjustable by the attribute :attr:`Matrix.abs_tol`.

        """
        if not isinstance(other, Matrix):
            raise TypeError("Matrix class required.")
        if self.shape != other.shape:
            raise TypeError("Matrices have different shapes.")
        return bool(np.all(np.isclose(self.matrix, other.matrix, atol=self.abs_tol)))

    def __mul__(self, other: Matrix | float) -> Matrix:
        """Matrix multiplication by another matrix or a float, returns a new
        matrix.

        """
        if isinstance(other, Matrix):
            return Matrix(matrix=np.matmul(self.matrix, other.matrix))
        else:
            matrix = Matrix(matrix=self.matrix * float(other))  # type: ignore
        return matrix

    __imul__ = __mul__

    def __add__(self, other: Matrix | float) -> Matrix:
        """Matrix addition by another matrix or a float, returns a new matrix."""
        if isinstance(other, Matrix):
            return Matrix(matrix=self.matrix + other.matrix)  # type: ignore
        else:
            return Matrix(matrix=self.matrix + float(other))  # type: ignore

    __iadd__ = __add__

    def __sub__(self, other: Matrix | float) -> Matrix:
        """Matrix subtraction by another matrix or a float, returns a new
        matrix.

        """
        if isinstance(other, Matrix):
            return Matrix(matrix=self.matrix - other.matrix)  # type: ignore
        else:
            return Matrix(matrix=self.matrix - float(other))  # type: ignore

    __isub__ = __sub__

    def transpose(self) -> Matrix:
        """Returns a new transposed matrix."""
        return Matrix(matrix=self.matrix.T)

    def inverse(self) -> Matrix:
        """Returns inverse of matrix as new object."""
        if self.nrows != self.ncols:
            raise TypeError("Inverse of non-square matrix not supported.")
        try:
            return Matrix(matrix=np.linalg.inv(self.matrix))  # type: ignore
        except np.linalg.LinAlgError:
            raise ZeroDivisionError

    def determinant(self) -> float:
        """Returns determinant of matrix, raises :class:`ZeroDivisionError`
        if matrix is singular.

        """
        return float(np.linalg.det(self.matrix))


class Solver(abc.ABC):
    @abc.abstractmethod
    def solve_matrix(self, B: MatrixData | NDArray) -> Matrix:
        ...

    @abc.abstractmethod
    def solve_vector(self, B: Iterable[float]) -> list[float]:
        ...


def quadratic_equation(a: float, b: float, c: float, abs_tol=1e-12) -> Sequence[float]:
    """Returns the solution for the quadratic equation ``a*x^2 + b*x + c = 0``.

    Returns 0-2 solutions as a tuple of floats.
    """
    if abs(a) < abs_tol:
        if abs(b) < abs_tol:
            return (-c,)
        return (-c / b,)
    try:
        discriminant = math.sqrt(b**2 - 4 * a * c)
    except ValueError:  # domain error, sqrt of a negative number
        return tuple()
    return ((-b + discriminant) / (2.0 * a)), ((-b - discriminant) / (2.0 * a))


# noinspection PyPep8Naming
def cubic_equation(a: float, b: float, c: float, d: float) -> Sequence[float]:
    """Returns the solution for the cubic equation ``a*x^3 + b*x^2 + c*x + d = 0``.

    Returns 0-3 solutions as a tuple of floats.
    """
    if abs(a) < 1e-12:
        try:
            return quadratic_equation(b, c, d)
        except ArithmeticError:  # complex solution
            return tuple()
    A = b / a
    B = c / a
    C = d / a
    AA = A * A
    A3 = A / 3.0

    Q = (3.0 * B - AA) / 9.0
    R = (9.0 * A * B - 27.0 * C - 2.0 * (AA * A)) / 54.0
    QQQ = Q * Q * Q
    D = QQQ + (R * R)  # polynomial discriminant

    if D >= 0.0:  # complex or duplicate roots
        sqrtD = math.sqrt(D)
        exp = 1.0 / 3.0
        S = math.copysign(1.0, R + sqrtD) * math.pow(abs(R + sqrtD), exp)
        T = math.copysign(1.0, R - sqrtD) * math.pow(abs(R - sqrtD), exp)
        ST = S + T
        if S - T:  # is complex
            return (-A3 + ST,)  # real root
        else:
            ST_2 = ST / 2.0
            return (
                -A3 + ST,  # real root
                -A3 - ST_2,  # real part of complex root
                -A3 - ST_2,  # real part of complex root
            )

    th = math.acos(R / math.sqrt(-QQQ))
    sqrtQ2 = math.sqrt(-Q) * 2.0
    return (
        sqrtQ2 * math.cos(th / 3.0) - A3,
        sqrtQ2 * math.cos((th + 2.0 * math.pi) / 3.0) - A3,
        sqrtQ2 * math.cos((th + 4.0 * math.pi) / 3.0) - A3,
    )


def numpy_matrix_solver(A: MatrixData | NDArray, B: MatrixData | NDArray) -> Matrix:
    """Solves the linear equation system given by a nxn Matrix A . x = B by the
    numpy.linalg.solve() function.

    Args:
        A: matrix [[a11, a12, ..., a1n], [a21, a22, ..., a2n], ... [an1, an2, ..., ann]]
        B: matrix [[b11, b12, ..., b1m], [b21, b22, ..., b2m], ... [bn1, bn2, ..., bnm]]

    Raises:
        numpy.linalg.LinAlgError: singular matrix

    """
    mat_A = np.array(A, dtype=np.float64)
    mat_B = np.array(B, dtype=np.float64)
    return Matrix(matrix=np.linalg.solve(mat_A, mat_B))  # type: ignore


def numpy_vector_solver(A: MatrixData | NDArray, B: Iterable[float]) -> list[float]:
    """Solves the linear equation system given by a nxn Matrix A . x = B,
    right-hand side quantities as vector B with n elements by the numpy.linalg.solve()
    function.

    Args:
        A: matrix [[a11, a12, ..., a1n], [a21, a22, ..., a2n], ... [an1, an2, ..., ann]]
        B: vector [b1, b2, ..., bn]

    Raises:
        numpy.linalg.LinAlgError: singular matrix

    """
    mat_A = np.array(A, dtype=np.float64)
    mat_B = np.array([[float(v)] for v in B], dtype=np.float64)
    return list(np.ravel(np.linalg.solve(mat_A, mat_B)))


class NumpySolver(Solver):
    """Replaces in v1.2 the :class:`LUDecomposition` solver."""

    def __init__(self, A: MatrixData | NDArray) -> None:
        self.mat_A = np.array(A, dtype=np.float64)

    def solve_matrix(self, B: MatrixData | NDArray) -> Matrix:
        """
        Solves the linear equation system given by the nxn Matrix
        A . x = B, right-hand side quantities as nxm Matrix B.

        Args:
            B: matrix [[b11, b12, ..., b1m], [b21, b22, ..., b2m],
                ... [bn1, bn2, ..., bnm]]

        Raises:
            numpy.linalg.LinAlgError: singular matrix

        """
        mat_B = np.array(B, dtype=np.float64)
        return Matrix(matrix=np.linalg.solve(self.mat_A, mat_B))  # type: ignore

    def solve_vector(self, B: Iterable[float]) -> list[float]:
        """Solves the linear equation system given by the nxn Matrix
        A . x = B, right-hand side quantities as vector B with n elements.

        Args:
            B: vector [b1, b2, ..., bn]

        Raises:
            numpy.linalg.LinAlgError: singular matrix

        """
        mat_B = np.array([[float(v)] for v in B], dtype=np.float64)
        return list(np.ravel(np.linalg.solve(self.mat_A, mat_B)))


def tridiagonal_vector_solver(A: MatrixData, B: Iterable[float]) -> list[float]:
    """Solves the linear equation system given by a tri-diagonal nxn Matrix
    A . x = B, right-hand side quantities as vector B. Matrix A is diagonal
    matrix defined by 3 diagonals [-1 (a), 0 (b), +1 (c)].

    Note: a0 is not used but has to be present, cn-1 is also not used and must
    not be present.

    If an :class:`ZeroDivisionError` exception occurs, the equation system can
    possibly be solved by :code:`BandedMatrixLU(A, 1, 1).solve_vector(B)`

    Args:
        A: diagonal matrix [[a0..an-1], [b0..bn-1], [c0..cn-1]] ::

            [[b0, c0, 0, 0, ...],
            [a1, b1, c1, 0, ...],
            [0, a2, b2, c2, ...],
            ... ]

        B: iterable of floats [[b1, b1, ..., bn]

    Returns:
        list of floats

    Raises:
        ZeroDivisionError: singular matrix

    """
    a, b, c = [list(v) for v in A]
    return _solve_tridiagonal_matrix(a, b, c, list(B))


def tridiagonal_matrix_solver(
    A: MatrixData | NDArray, B: MatrixData | NDArray
) -> Matrix:
    """Solves the linear equation system given by a tri-diagonal nxn Matrix
    A . x = B, right-hand side quantities as nxm Matrix B. Matrix A is diagonal
    matrix defined by 3 diagonals [-1 (a), 0 (b), +1 (c)].

    Note: a0 is not used but has to be present, cn-1 is also not used and must
    not be present.

    If an :class:`ZeroDivisionError` exception occurs, the equation system
    can possibly be solved by :code:`BandedMatrixLU(A, 1, 1).solve_vector(B)`

    Args:
        A: diagonal matrix [[a0..an-1], [b0..bn-1], [c0..cn-1]] ::

            [[b0, c0, 0, 0, ...],
            [a1, b1, c1, 0, ...],
            [0, a2, b2, c2, ...],
            ... ]

        B: matrix [[b11, b12, ..., b1m],
                   [b21, b22, ..., b2m],
                   ...
                   [bn1, bn2, ..., bnm]]

    Returns:
        matrix as :class:`Matrix` object

    Raises:
        ZeroDivisionError: singular matrix

    """
    a, b, c = [list(v) for v in A]
    if not isinstance(B, Matrix):
        matrix_b = Matrix(matrix=[list(row) for row in B])
    else:
        matrix_b = cast(Matrix, B)
    if matrix_b.nrows != len(b):
        raise ValueError("Row count of matrices A and B has to match.")

    return Matrix(
        matrix=[_solve_tridiagonal_matrix(a, b, c, col) for col in matrix_b.cols()]
    ).transpose()


def _solve_tridiagonal_matrix(
    a: list[float], b: list[float], c: list[float], r: list[float]
) -> list[float]:
    """Solves the linear equation system given by a tri-diagonal
    Matrix(a, b, c) . x = r.

    Matrix configuration::

        [[b0, c0, 0, 0, ...],
        [a1, b1, c1, 0, ...],
        [0, a2, b2, c2, ...],
        ... ]

    Args:
        a: lower diagonal [a0 .. an-1], a0 is not used but has to be present
        b: central diagonal [b0 .. bn-1]
        c: upper diagonal [c0 .. cn-1], cn-1 is not used and must not be present
        r: right-hand side quantities

    Returns:
        vector x as list of floats

    Raises:
        ZeroDivisionError: singular matrix

    """
    n: int = len(a)
    u: list[float] = [0.0] * n
    gam: list[float] = [0.0] * n
    bet: float = b[0]
    u[0] = r[0] / bet
    for j in range(1, n):
        gam[j] = c[j - 1] / bet
        bet = b[j] - a[j] * gam[j]
        u[j] = (r[j] - a[j] * u[j - 1]) / bet

    for j in range((n - 2), -1, -1):
        u[j] -= gam[j + 1] * u[j + 1]
    return u


def banded_matrix(A: Matrix, check_all=True) -> tuple[Matrix, int, int]:
    """Transform matrix A into a compact banded matrix representation.
    Returns compact representation as :class:`Matrix` object and
    lower- and upper band count m1 and m2.

    Args:
        A: input :class:`Matrix`
        check_all: check all diagonals if ``True`` or abort testing
            after first all zero diagonal if ``False``.

    """
    m1, m2 = detect_banded_matrix(A, check_all)
    m = compact_banded_matrix(A, m1, m2)
    return m, m1, m2


def detect_banded_matrix(A: Matrix, check_all=True) -> tuple[int, int]:
    """Returns lower- and upper band count m1 and m2.

    Args:
        A: input :class:`Matrix`
        check_all: check all diagonals if ``True`` or abort testing
            after first all zero diagonal if ``False``.

    """

    def detect_m2() -> int:
        m2: int = 0
        for d in range(1, A.ncols):
            if any(A.diag(d)):
                m2 = d
            elif not check_all:
                break
        return m2

    def detect_m1() -> int:
        m1: int = 0
        for d in range(1, A.nrows):
            if any(A.diag(-d)):
                m1 = d
            elif not check_all:
                break
        return m1

    return detect_m1(), detect_m2()


def compact_banded_matrix(A: Matrix, m1: int, m2: int) -> Matrix:
    """Returns compact banded matrix representation as :class:`Matrix` object.

    Args:
        A: matrix to transform
        m1: lower band count, excluding main matrix diagonal
        m2: upper band count, excluding main matrix diagonal

    """
    if A.nrows != A.ncols:
        raise TypeError("Square matrix required.")

    m = Matrix()

    for d in range(m1, 0, -1):
        col = [0.0] * d
        col.extend(A.diag(-d))
        m.append_col(col)

    m.append_col(A.diag(0))

    for d in range(1, m2 + 1):
        col = A.diag(d)
        col.extend([0.0] * d)
        m.append_col(col)
    return m


class BandedMatrixLU(Solver):
    """Represents a LU decomposition of a compact banded matrix."""

    def __init__(self, A: Matrix, m1: int, m2: int):
        lu_decompose = _lu_decompose
        if USE_C_EXT:
            # import error shows an installation issue
            from ezdxf.acc.np_support import lu_decompose  # type: ignore

        self.m1: int = int(m1)
        self.m2: int = int(m2)
        self.upper, self.lower, self.index = lu_decompose(A.matrix, self.m1, self.m2)

    @property
    def nrows(self) -> int:
        """Count of matrix rows."""
        return self.upper.shape[0]

    def solve_vector(self, B: Iterable[float]) -> list[float]:
        """Solves the linear equation system given by the banded nxn Matrix
        A . x = B, right-hand side quantities as vector B with n elements.

        Args:
            B: vector [b1, b2, ..., bn]

        Returns:
            vector as list of floats

        """

        solve_vector_banded_matrix = _solve_vector_banded_matrix
        if USE_C_EXT:
            # import error shows an installation issue
            from ezdxf.acc.np_support import solve_vector_banded_matrix  # type: ignore

        x: NDArray = np.array(B, dtype=np.float64)
        if len(x) != self.nrows:
            raise ValueError(
                "Item count of vector B has to be equal to matrix row count."
            )
        return list(
            solve_vector_banded_matrix(
                x, self.upper, self.lower, self.index, self.m1, self.m2
            )
        )

    def solve_matrix(self, B: MatrixData | NDArray) -> Matrix:
        """
        Solves the linear equation system given by the banded nxn Matrix
        A . x = B, right-hand side quantities as nxm Matrix B.

        Args:
            B: matrix [[b11, b12, ..., b1m], [b21, b22, ..., b2m],
                ... [bn1, bn2, ..., bnm]]

        Returns:
            matrix as :class:`Matrix` object

        """
        matrix_b = Matrix(matrix=B)
        if matrix_b.nrows != self.nrows:
            raise ValueError("Row count of self and matrix B has to match.")

        return Matrix(
            matrix=[self.solve_vector(col) for col in matrix_b.cols()]
        ).transpose()


def _lu_decompose(A: NDArray, m1: int, m2: int) -> tuple[NDArray, NDArray, NDArray]:
    # upper triangle of LU decomposition
    upper: NDArray = np.array(A, dtype=np.float64)

    # lower triangle of LU decomposition
    n: int = upper.shape[0]
    lower: NDArray = np.zeros((n, m1), dtype=np.float64)
    index: NDArray = np.zeros((n,), dtype=np.int64)

    mm: int = m1 + m2 + 1
    l: int = m1
    for i in range(m1):
        for j in range(m1 - i, mm):
            upper[i][j - l] = upper[i][j]
        l -= 1
        for j in range(mm - l - 1, mm):
            upper[i][j] = 0.0

    l = m1
    for k in range(n):
        dum = upper[k][0]
        i = k
        if l < n:
            l += 1
        for j in range(k + 1, l):
            if abs(upper[j][0]) > abs(dum):
                dum = upper[j][0]
                i = j
        index[k] = i + 1
        if i != k:
            for j in range(mm):
                upper[k][j], upper[i][j] = upper[i][j], upper[k][j]

        for i in range(k + 1, l):
            dum = upper[i][0] / upper[k][0]
            lower[k][i - k - 1] = dum
            for j in range(1, mm):
                upper[i][j - 1] = upper[i][j] - dum * upper[k][j]
            upper[i][mm - 1] = 0.0
    return upper, lower, index


def _solve_vector_banded_matrix(
    x: NDArray,
    upper: NDArray,
    lower: NDArray,
    index: NDArray,
    m1: int,
    m2: int,
) -> NDArray:
    """Solves the linear equation system given by the banded nxn Matrix
    A . x = B, right-hand side quantities as vector B with n elements.

    Args:
        B: vector [b1, b2, ..., bn]

    Returns:
        vector as list of floats

    """
    n: int = upper.shape[0]
    if x.shape[0] != n:
        raise ValueError("Item count of vector B has to be equal to matrix row count.")

    al: NDArray = lower
    au: NDArray = upper
    mm: int = m1 + m2 + 1
    l: int = m1
    for k in range(n):
        j = index[k] - 1
        if j != k:
            x[k], x[j] = x[j], x[k]
        if l < n:
            l += 1
        for j in range(k + 1, l):
            x[j] -= al[k][j - k - 1] * x[k]

    l = 1
    for i in range(n - 1, -1, -1):
        dum = x[i]
        for k in range(1, l):
            dum -= au[i][k] * x[k + i]
        x[i] = dum / au[i][0]
        if l < mm:
            l += 1

    return x



# Copyright (c) 2010-2024, Manfred Moitzi
# License: MIT License
import pytest
from math import radians
import numpy as np

from ezdxf.math import xround, Vec2
from ezdxf.math.construct2d import *


def test_left_of_line():
    assert is_point_left_of_line(Vec2(-1, 0), Vec2(0, 0), Vec2(0.1, 1)) is True
    assert is_point_left_of_line(Vec2(1, 0), Vec2(0, 0), Vec2(0, -1)) is True
    assert is_point_left_of_line(Vec2(-1, -1), Vec2(0, 0), Vec2(0.1, 1)) is True


def test_point_to_line_relation_left():
    assert point_to_line_relation(Vec2(-1, 0), Vec2(0, 0), Vec2(0.1, 1)) == -1
    assert point_to_line_relation(Vec2(1, 0), Vec2(0, 0), Vec2(0, -1)) == -1
    assert point_to_line_relation(Vec2(-1, -1), Vec2(0, 0), Vec2(0.1, 1)) == -1


def test_left_of_line_or_on_the_line():
    # vertical line
    assert (
        is_point_left_of_line(Vec2(1, 0), Vec2(0, 0), Vec2(0, 1), colinear=True)
        is False
    )
    assert (
        is_point_left_of_line(
            Vec2(0, 0.5), Vec2(0, 0), Vec2(0, 1), colinear=True
        )
        is True
    )
    assert (
        is_point_left_of_line(
            Vec2(-1, 0.5), Vec2(0, 0), Vec2(0, 1), colinear=True
        )
        is True
    )
    # horizontal line
    assert (
        is_point_left_of_line(Vec2(0, 1), Vec2(0, 0), Vec2(1, 0), colinear=True)
        is True
    )
    assert (
        is_point_left_of_line(Vec2(0, 0), Vec2(0, 0), Vec2(1, 0), colinear=True)
        is True
    )
    assert (
        is_point_left_of_line(
            Vec2(0, -1), Vec2(0, 0), Vec2(1, 0), colinear=True
        )
        is False
    )
    # 45 deg line
    assert (
        is_point_left_of_line(Vec2(0, 0), Vec2(0, 0), Vec2(1, 1), colinear=True)
        is True
    )
    assert (
        is_point_left_of_line(
            Vec2(0.5, 0.5), Vec2(0, 0), Vec2(1, 1), colinear=True
        )
        is True
    )
    assert (
        is_point_left_of_line(Vec2(1, 1), Vec2(0, 0), Vec2(1, 1), colinear=True)
        is True
    )
    assert (
        is_point_left_of_line(
            Vec2(0.5, 0.49), Vec2(0, 0), Vec2(1, 1), colinear=True
        )
        is False
    )


def test_point_ot_line_relation_on_line():
    # vertical line
    assert point_to_line_relation(Vec2(0, 2), Vec2(0, 0), Vec2(0, 1)) == 0
    assert point_to_line_relation(Vec2(0, -1), Vec2(0, 0), Vec2(0, 1)) == 0
    assert point_to_line_relation(Vec2(0, 0.5), Vec2(0, 0), Vec2(0, 1)) == 0
    # horizontal line
    assert point_to_line_relation(Vec2(1, 0), Vec2(0, 0), Vec2(1, 0)) == 0
    assert point_to_line_relation(Vec2(-1, 0), Vec2(0, 0), Vec2(1, 0)) == 0
    assert point_to_line_relation(Vec2(0, 0), Vec2(0, 0), Vec2(1, 0)) == 0
    # 45 deg line
    assert point_to_line_relation(Vec2(0, 0), Vec2(0, 0), Vec2(1, 1)) == 0
    assert point_to_line_relation(Vec2(0.5, 0.5), Vec2(0, 0), Vec2(1, 1)) == 0
    assert point_to_line_relation(Vec2(1, 1), Vec2(0, 0), Vec2(1, 1)) == 0
    assert point_to_line_relation(Vec2(-0.5, -0.5), Vec2(0, 0), Vec2(1, 1)) == 0


def test_xround():
    assert xround(9.999, 0.0) == 10
    assert xround(9.999, 0.5) == 10
    assert xround(9.75, 0.5) == 10
    assert xround(9.74, 0.5) == 9.5
    assert xround(9.74, 0.25) == 9.75
    assert xround(9.626, 0.25) == 9.75
    assert xround(9.624, 0.25) == 9.50
    assert xround(9.624, 0.1) == 9.6
    assert xround(9.624, 0.01) == 9.62
    assert xround(9.626, 0.01) == 9.63
    assert xround(9.626, 0.05) == 9.65

    assert xround(19.1, 1.0) == 19
    assert xround(19.1, 2.0) == 20
    assert xround(18.9, 2.0) == 18
    assert xround(18.9, 5.0) == 20
    assert xround(18.9, 10) == 20
    assert xround(1234, 10) == 1230
    assert xround(1236, 10) == 1240


def test_enclosing_angles():
    assert (
        enclosing_angles(
            radians(45),
            start_angle=radians(45),
            end_angle=radians(45),
            ccw=True,
        )
        is True
    )
    assert (
        enclosing_angles(
            radians(45),
            start_angle=radians(45),
            end_angle=radians(45),
            ccw=False,
        )
        is True
    )

    assert (
        enclosing_angles(
            radians(90),
            start_angle=radians(45),
            end_angle=radians(135),
            ccw=True,
        )
        is True
    )
    assert (
        enclosing_angles(
            radians(90),
            start_angle=radians(45),
            end_angle=radians(135),
            ccw=False,
        )
        is False
    )

    assert (
        enclosing_angles(
            radians(0),
            start_angle=radians(45),
            end_angle=radians(135),
            ccw=True,
        )
        is False
    )
    assert (
        enclosing_angles(
            radians(0),
            start_angle=radians(45),
            end_angle=radians(135),
            ccw=False,
        )
        is True
    )

    assert (
        enclosing_angles(
            radians(45),
            start_angle=radians(50),
            end_angle=radians(40),
            ccw=True,
        )
        is False
    )
    assert (
        enclosing_angles(
            radians(45),
            start_angle=radians(50),
            end_angle=radians(40),
            ccw=False,
        )
        is True
    )

    assert (
        enclosing_angles(
            radians(90),
            start_angle=radians(135),
            end_angle=radians(45),
            ccw=True,
        )
        is False
    )
    assert (
        enclosing_angles(
            radians(90),
            start_angle=radians(135),
            end_angle=radians(45),
            ccw=False,
        )
        is True
    )

    assert (
        enclosing_angles(
            radians(270),
            start_angle=radians(135),
            end_angle=radians(45),
            ccw=True,
        )
        is True
    )
    assert (
        enclosing_angles(
            radians(270),
            start_angle=radians(135),
            end_angle=radians(45),
            ccw=False,
        )
        is False
    )


def test_no_points():
    assert closest_point((0, 0), []) is None


def test_one_points():
    assert closest_point((0, 0), [(1, 1)]) == (1, 1)


def test_two_points():
    assert closest_point((0, 0), [(0, 0, 1), (1, 1, 1)]) == (0, 0, 1)


def test_more_points():
    assert closest_point(
        (0, 0), [(0, 0, 1), (1, 1, 1), (2, 2, 2), (0, 0, -0.5)]
    ) == (0, 0, -0.5)


def test_decdeg2dms():
    ss = 1.0 / 3600.0
    mm = 1.0 / 60.0
    for value in np.linspace(-2, 2, 71):
        d, m, s = decdeg2dms(value)
        result = d + m * mm + s * ss
        assert value == pytest.approx(result)
    assert decdeg2dms(-1) == (-1, 0, 0)
    assert decdeg2dms(0) == (0, 0, 0)
    assert decdeg2dms(1) == (1, 0, 0)


def test_linspace():
    assert list(np.linspace(1, 4, num=4)) == [1, 2, 3, 4]
    assert list(np.linspace(1, 4, num=1)) == [1]
    assert list(np.linspace(1, 4, num=0)) == []
    assert list(np.linspace(1, 5, num=4, endpoint=False)) == [1, 2, 3, 4]
    assert list(np.linspace(2, -2, num=5)) == [2, 1, 0, -1, -2]
    with pytest.raises(ValueError):
        list(np.linspace(1, 4, num=-1))


def test_area():
    assert area([(4, 6), (4, -4), (8, -4), (8, -8), (-4, -8), (-4, 6)]) == 128
